"""
This module handles registering multimae models into timm.
"""
import logging
import torch
import numpy as np
from pathlib import Path
from functools import partial
from timm.models import FeatureInfo
from timm.models._builder import build_model_with_cfg
from timm.models._registry import generate_default_cfgs, register_model

from terratorch.datasets.utils import HLSBands, Modalities, S1Bands, DEMBands, LULCclasses
from terratorch.models.backbones.multimae.multimae import MultiMAE, MultiViT
from terratorch.models.backbones.multimae.criterion import MaskedMSELoss, MaskedCrossEntropyLoss
from terratorch.models.backbones.multimae.input_adapters import PatchedInputAdapter, SemSegInputAdapter
from terratorch.models.backbones.multimae.output_adapters import SpatialOutputAdapter, ConvNeXtAdapter

logger = logging.getLogger(__name__)

def _cfg(file: Path = "", **kwargs) -> dict:
    return {
        "file": file,
        "source": "file",
        "license": "mit",
        **kwargs,
    }


# TODO: Add pretrained models
# PRETRAINED_BANDS: list[HLSBands | int] = [
#     HLSBands.BLUE,
#     HLSBands.GREEN,
#     HLSBands.RED,
# ]

# default_cfgs = generate_default_cfgs(
#     {
#         "multimae_base": _cfg(
#             file=""
#         ),
#         "multimae_large": _cfg(
#             file=""
#         )
#
#     }
# )


# TODO: make these user definable
DOMAIN_CONF = {
    Modalities.S1: {
        "channels": 2,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=2),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=2),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.S2L1C: {
        "channels": 13,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=13),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=13),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.S2L2A: {
        "channels": 12,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=12),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=12),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.S2RGB: {
        "channels": 3,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=3),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=3),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.DEM: {
        "channels": 1,
        "stride_level": 1,
        "input_adapter": partial(PatchedInputAdapter, num_channels=1),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=1),
        "loss": MaskedMSELoss,
        "image_size": 224,
        "patch_size": 16,
    },
    Modalities.LULC: {
        "classes": 9,
        "stride_level": 1,
        "input_adapter": partial(SemSegInputAdapter, num_classes=9),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=9),  # Used in pretraining
        "loss": partial(MaskedCrossEntropyLoss, label_smoothing=0.0),
        "image_size": 224,
        "patch_size": 16,
    },
    'segmentation': {  # TODO: Test generalized semseg head from MultiMAE!
        "classes": 9,
        "stride_level": 1,
        "input_adapter": partial(SemSegInputAdapter, num_classes=9),
        "output_adapter": partial(SpatialOutputAdapter, num_channels=9),  # Used in pretraining
        "loss": partial(MaskedCrossEntropyLoss, label_smoothing=0.0),
        "image_size": 224,
        "patch_size": 16,
    },
}


def _instantiate_input_adapter_from_dict(spec: dict) -> PatchedInputAdapter | SemSegInputAdapter:
    return spec["input_adapter"](
        stride_level=spec["stride_level"],
        patch_size_full=spec["patch_size"],
        image_size=spec["image_size"],
    )


def _parse_input_adapters(
    adapter_spec: list | dict[str, str | dict[str, int | str]],
) -> dict[str, PatchedInputAdapter | SemSegInputAdapter]:

    if isinstance(adapter_spec, list):
        # list to dict
        adapter_spec = {m: m for m in adapter_spec}
    if isinstance(adapter_spec, dict) and len(set(adapter_spec.keys())) != len(adapter_spec.keys()):
        msg = "Duplicate keys in input adapters"
        raise Exception(msg)
    input_adapters = {}

    for adapter_name, spec in adapter_spec.items():
        match spec:
            case str(spec):
                try:
                    spec = Modalities(spec.upper())
                except ValueError:
                    pass

                if spec in DOMAIN_CONF.keys():
                    input_adapters[adapter_name] = _instantiate_input_adapter_from_dict(DOMAIN_CONF[spec])
                else:
                    msg = f"Input Domain {adapter_name} does not exist. Choose one of {list(DOMAIN_CONF.keys())}"
                    raise ValueError(msg)
            case {"type": "PatchedInputAdapter", "num_channels": num_channels, **kwargs}:
                input_adapters[adapter_name] = PatchedInputAdapter(num_channels=num_channels, **kwargs)
            case {"type": "SemSegInputAdapter", "num_classes": num_classes,  **kwargs}:
                input_adapters[adapter_name] = SemSegInputAdapter(num_classes=num_classes, **kwargs)
            case _:
                msg = f"Invalid input adapter config for adapter {adapter_name}"
                raise ValueError(msg)
    return input_adapters


def _instantiate_output_adapter_from_dict(spec: dict, task: str, context_tasks: list[str], # num_channels: int
                                          ) -> SpatialOutputAdapter | ConvNeXtAdapter:
    return spec["output_adapter"](
        stride_level=spec["stride_level"],
        patch_size_full=spec["patch_size"],
        image_size=spec["image_size"],
        task=task,
        context_tasks=context_tasks,
        # num_channels=spec['channels'],  # TODO: Not passed in pretraining code
    )


def _parse_output_adapters(
    adapter_spec: list | dict[str, str | dict[str, int | str]],
) -> dict[str, SpatialOutputAdapter | SpatialOutputAdapter]:

    if isinstance(adapter_spec, list):
        # list to dict
        adapter_spec = {m: m for m in adapter_spec}
    if isinstance(adapter_spec, dict) and len(set(adapter_spec.keys())) != len(adapter_spec.keys()):
        msg = "Duplicate keys in output adapters"
        raise Exception(msg)
    output_adapters = {}

    for adapter_name, spec in adapter_spec.items():
        match spec:
            case str(spec):
                try:
                    spec = Modalities(spec.upper())
                except ValueError:
                    pass

                if spec in DOMAIN_CONF.keys():
                    output_adapters[adapter_name] = _instantiate_output_adapter_from_dict(
                        DOMAIN_CONF[spec],
                        task=adapter_name,
                        context_tasks=list(adapter_spec.keys()),
                    )
                else:
                    msg = f"output Domain {adapter_name} does not exist. Choose one of {list(DOMAIN_CONF.keys())}"
                    raise ValueError(msg)
            case {"type": "SpatialOutputAdapter", "num_channels": num_channels, **kwargs}:  # Used for pre-training
                output_adapters[adapter_name] = SpatialOutputAdapter(
                    task=adapter_name,
                    context_tasks=list(adapter_spec.keys()),
                    num_channels=num_channels,
                    **kwargs
                )
            case {"type": "ConvNeXtAdapter", "num_classes": num_classes,  **kwargs}:
                output_adapters[adapter_name] = ConvNeXtAdapter(num_classes=num_classes, **kwargs)
            case _:
                msg = f"Invalid output adapter config for adapter {adapter_name}"
                raise ValueError(msg)
    return output_adapters


# If you need to adapt the checkpoint file, do it here
def checkpoint_filter_fn(
    modalities: list[str],
    state_dict: dict[str, torch.Tensor],
    model: torch.nn.Module,
):
    new_state_dict = {}

    for k, v in state_dict.items():
        if "output_adapters" in k:
            continue

        # drop pos emb
        if "pos_emb" in k:
            continue

        if k.startswith("input_adapters."):
            try:
                modality_name = k.split(".")[1]
                modality = Modalities(modality_name)
            except ValueError:
                print(f"Modality {modality_name} is not in allowed modalities. Skipping {k}.")
                continue
            if modality.value not in modalities:
                print(f"Removing input adapter for {modality_name}: {k}")
                continue
        if k.startswith("encoder."):
            new_k = "layers." + k.removeprefix("encoder.")
            new_state_dict[new_k] = v
        else:
            new_state_dict[k] = v
    return new_state_dict


class PrepareMultimodalFeaturesForDecoder:
    def __init__(self, modalities: list[str], merging_method: str = 'concat'):
        self.modalities = modalities
        self.merging_method = merging_method

    def __call__(self, x: list[torch.Tensor]) -> list[torch.Tensor]:
        if len(x) == 2:
            # MultiMAE decoder was used. Return predictions of first modality.
            preds = list(x[0].values())
            assert len(preds) != 1, "Terratorch can only handle one output modality."
            return preds[0]

        for output_index in range(len(x)):
            x[output_index] = x[output_index].permute(0, 2, 1)
            x[output_index] = x[output_index][:, :, :-1]  # remove global token
            img_shape = int(np.sqrt(x[output_index].shape[-1] / len(self.modalities)))
            if self.merging_method == 'concat':
                x[output_index] = x[output_index].reshape(x[output_index].shape[0], -1, img_shape, img_shape)
            else:
                raise ValueError(f"Unsupported merging method {self.merging_method}")
                # TODO: Implement other methods, move to forward?

        return x


def _create_multimae(
    variant: str,
    modalities: list[str],
    pretrained: bool = False,
    features_only: bool = True,
    merging_method: str = None,
    **kwargs,
):
    model: torch.nn.Module = build_model_with_cfg(
        MultiViT if features_only else MultiMAE,  # MultiViT is an encoder-only model
        variant,
        pretrained,
        # if you need to adapt the checkpoint file
        pretrained_filter_fn=partial(checkpoint_filter_fn, modalities),
        pretrained_strict=False,
        feature_cfg={
            "flatten_sequential": True,
        },
        features_only=False,
        merging_method=merging_method,
        **kwargs,
    )
    default_out_indices = list(range(len(model.layers)))
    out_indices = kwargs.get("out_indices", default_out_indices)
    model.feature_info = FeatureInfo(model.feature_info, out_indices)

    model.prepare_features_for_image_model = PrepareMultimodalFeaturesForDecoder(
        modalities,
        merging_method=merging_method
    )
    return model


@register_model
def multimae_base(
    input_adapters: dict[str, str | dict[str, int | str]] | None = None,
    output_adapters: dict[str, str | dict[str, int | str]] | None = None,
    pretrained: bool = False,
    features_only: bool = True,
    **kwargs,
) -> torch.nn.Module:
    """MultiMAE base model."""

    if input_adapters is None:
        input_adapters = ['S1', 'S2L1C', 'S2L2A', 'DEM', 'LULC']
        logger.warning(f'Using default adapters.')
    input_adapters = _parse_input_adapters(input_adapters)

    if output_adapters is not None:
        output_adapters = _parse_output_adapters(output_adapters)

    model_args = {
        "input_adapters": input_adapters,
        "output_adapters": output_adapters,
        "dim_tokens": 768,
        "depth": 12,
        "num_heads": 12,
        "mlp_ratio": 4,
        "qkv_bias": True,
        "norm_layer": partial(torch.nn.LayerNorm, eps=1e-6),
    }

    kwargs.pop('features_only', None)
    merging_method = None if output_adapters else kwargs.get('merging_method', 'concat')

    transformer = _create_multimae(
        "multimae_base",
        list(input_adapters.keys()),
        pretrained=pretrained,
        features_only=output_adapters is None,
        merging_method=merging_method,
        **dict(model_args, **kwargs),
    )
    return transformer


@register_model
def multimae_large(
    input_adapters: dict[str, str | dict[str, int | str]] | None = None,
    output_adapters: dict[str, str | dict[str, int | str]] | None = None,
    pretrained: bool = False,  # noqa: FBT002, FBT001
    **kwargs,
) -> torch.nn.Module:
    """MultiMAE large model."""

    if input_adapters is None:
        input_adapters = ['S1', 'S2L1C', 'S2L2A', 'DEM', 'LULC']
    input_adapters = _parse_input_adapters(input_adapters)

    if output_adapters is not None:
        output_adapters = _parse_output_adapters(output_adapters)

    model_args = {
        "input_adapters": input_adapters,
        "output_adapters": output_adapters,
        "dim_tokens": 1024,
        "depth": 24,
        "num_heads": 16,
        "mlp_ratio": 4,
        "qkv_bias": True,
        "norm_layer": partial(torch.nn.LayerNorm, eps=1e-6),
    }

    kwargs.pop('features_only', None)
    merging_method = None if output_adapters else kwargs.get('merging_method', 'concat')

    transformer = _create_multimae(
        "multimae_large",
        list(input_adapters.keys()),
        pretrained=pretrained,
        features_only=output_adapters is None,
        merging_method=merging_method,
        **dict(model_args, **kwargs),
    )
    return transformer



from torchgeo.trainers import BaseTask
import torch.nn as nn
import torch
import logging
logger = logging.getLogger(__name__)

from terratorch.registry import MODEL_FACTORY_REGISTRY

class WxCTask(BaseTask):
    def __init__(self, model_factory, model_args: dict, mode:str='train', learning_rate=0.1):
        if mode not in ['train', 'eval']:
            raise ValueError(f'mode {mode} is not supported. (train, eval)')
        self.model_args = model_args

        self.model_factory = MODEL_FACTORY_REGISTRY.build(model_factory)

        self.learning_rate = learning_rate
        super().__init__()

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters(), lr=self.learning_rate)
    
    def configure_models(self):
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.model = self.model_factory.build_model(**self.model_args)
        self.model = self.model.to(device)
        layer_devices = []
        for name, module in self.model.named_children():
            device = next(module.parameters(), torch.tensor([])).device
            layer_devices.append((name, str(device)))
        logging.debug(layer_devices)

    def training_step(self, batch, batch_idx):
        output: torch.Tensor = self.model(batch)

    def train_dataloader(self):
        return DataLoader(self.dataset, batch_size=self.batch_size, shuffle=True)
        



import torch
import torch.nn.functional as F  # noqa: N812
from torch import Tensor, nn
import warnings

from terratorch.registry import TERRATORCH_DECODER_REGISTRY
from .utils import ConvModule


# Adapted from MMSegmentation
@TERRATORCH_DECODER_REGISTRY.register
class UperNetDecoder(nn.Module):
    """UperNetDecoder. Adapted from MMSegmentation."""

    def __init__(
        self,
        embed_dim: list[int],
        pool_scales: tuple[int] = (1, 2, 3, 6),
        channels: int = 256,
        align_corners: bool = True,  # noqa: FBT001, FBT002
        scale_modules: bool = False,
    ):
        """Constructor

        Args:
            embed_dim (list[int]): Input embedding dimension for each input.
            pool_scales (tuple[int], optional): Pooling scales used in Pooling Pyramid
                Module applied on the last feature. Default: (1, 2, 3, 6).
            channels (int, optional): Channels used in the decoder. Defaults to 256.
            align_corners (bool, optional): Wheter to align corners in rescaling. Defaults to True.
            scale_modules (bool, optional): Whether to apply scale modules to the inputs. Needed for plain ViT.
                Defaults to False.
        """
        super().__init__()
        if scale_modules:
            # TODO: remove scale_modules before v1?
            warnings.warn(
                "DeprecationWarning: scale_modules is deprecated and will be removed in future versions. "
                "Use LearnedInterpolateToPyramidal neck instead.",
                stacklevel=1,
            )

        self.scale_modules = scale_modules
        if scale_modules:
            self.fpn1 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim[0],
                                embed_dim[0] // 2, 2, 2),
                nn.BatchNorm2d(embed_dim[0] // 2),
                nn.GELU(),
                nn.ConvTranspose2d(embed_dim[0] // 2,
                                embed_dim[0] // 4, 2, 2))
            self.fpn2 = nn.Sequential(
                nn.ConvTranspose2d(embed_dim[1],
                                embed_dim[1] // 2, 2, 2))
            self.fpn3 = nn.Sequential(nn.Identity())
            self.fpn4 = nn.Sequential(nn.MaxPool2d(kernel_size=2, stride=2))
            self.embed_dim = [embed_dim[0] // 4, embed_dim[1] // 2, embed_dim[2], embed_dim[3]]
        else:
            self.embed_dim = embed_dim

        self.out_channels = channels
        self.channels = channels
        self.align_corners = align_corners
        # PSP Module
        self.psp_modules = PPM(
            pool_scales,
            self.embed_dim[-1],
            self.channels,
            align_corners=self.align_corners,
        )
        self.bottleneck = ConvModule(
            self.embed_dim[-1] + len(pool_scales) * self.channels, self.channels, 3, padding=1, inplace=True
        )
        # FPN Module
        self.lateral_convs = nn.ModuleList()
        self.fpn_convs = nn.ModuleList()
        for embed_dim in self.embed_dim[:-1]:  # skip the top layer
            l_conv = ConvModule(
                embed_dim,
                self.channels,
                1,
                inplace=False,
            )
            fpn_conv = ConvModule(
                self.channels,
                self.channels,
                3,
                padding=1,
                inplace=False,
            )
            self.lateral_convs.append(l_conv)
            self.fpn_convs.append(fpn_conv)

        self.fpn_bottleneck = ConvModule(len(self.embed_dim) * self.channels, self.channels, 3, padding=1, inplace=True)

    def psp_forward(self, inputs):
        """Forward function of PSP module."""
        x = inputs[-1]
        psp_outs = [x]
        psp_outs.extend(self.psp_modules(x))
        psp_outs = torch.cat(psp_outs, dim=1)
        output = self.bottleneck(psp_outs)

        return output

    def forward(self, inputs):
        """Forward function for feature maps before classifying each pixel with
        Args:
            inputs (list[Tensor]): List of multi-level img features.

        Returns:
            feats (Tensor): A tensor of shape (batch_size, self.channels,
                H, W) which is feature map for last layer of decoder head.
        """

        if self.scale_modules:
            scaled_inputs = []
            scaled_inputs.append(self.fpn1(inputs[0]))
            scaled_inputs.append(self.fpn2(inputs[1]))
            scaled_inputs.append(self.fpn3(inputs[2]))
            scaled_inputs.append(self.fpn4(inputs[3]))
            inputs = scaled_inputs
        # build laterals
        laterals = [lateral_conv(inputs[i]) for i, lateral_conv in enumerate(self.lateral_convs)]
        laterals.append(self.psp_forward(inputs))

        # build top-down path
        used_backbone_levels = len(laterals)
        for i in range(used_backbone_levels - 1, 0, -1):
            prev_shape = laterals[i - 1].shape[2:]
            laterals[i - 1] = laterals[i - 1] + torch.nn.functional.interpolate(
                laterals[i], size=prev_shape, mode="bilinear", align_corners=self.align_corners
            )

        # build outputs
        fpn_outs = [self.fpn_convs[i](laterals[i]) for i in range(used_backbone_levels - 1)]
        # append psp feature
        fpn_outs.append(laterals[-1])

        for i in range(used_backbone_levels - 1, 0, -1):
            fpn_outs[i] = torch.nn.functional.interpolate(
                fpn_outs[i], size=fpn_outs[0].shape[2:], mode="bilinear", align_corners=self.align_corners
            )
        fpn_outs = torch.cat(fpn_outs, dim=1)
        feats = self.fpn_bottleneck(fpn_outs)
        return feats


class PPM(nn.ModuleList):
    """Pooling Pyramid Module used in PSPNet."""

    def __init__(self, pool_scales, in_channels, channels, align_corners):
        """Constructor

        Args:
            pool_scales (tuple[int]): Pooling scales used in Pooling Pyramid
                Module.
            in_channels (int): Input channels.
            channels (int): Channels after modules, before conv_seg.
            align_corners (bool): align_corners argument of F.interpolate.
        """
        super().__init__()
        self.pool_scales = pool_scales
        self.align_corners = align_corners
        self.in_channels = in_channels
        self.channels = channels

        for pool_scale in pool_scales:
            self.append(
                nn.Sequential(
                    nn.AdaptiveAvgPool2d(pool_scale),
                    ConvModule(self.in_channels, self.channels, 1, inplace=True),
                )
            )

    def forward(self, x):
        """Forward function."""
        ppm_outs = []
        for ppm in self:
            ppm_out = ppm(x)
            upsampled_ppm_out = torch.nn.functional.interpolate(
                ppm_out, size=x.size()[2:], mode="bilinear", align_corners=self.align_corners
            )
            ppm_outs.append(upsampled_ppm_out)
        return ppm_outs



# Copyright contributors to the Terratorch project

import logging
import warnings

import torch
from torch import nn

from terratorch.datasets import HLSBands, OpticalBands, SARBands
import collections

def patch_embed_weights_are_compatible(model_patch_embed: torch.Tensor, checkpoint_patch_embed: torch.Tensor) -> bool:
    # check all dimensions are the same except for channel dimension
    if len(model_patch_embed.shape) != len(checkpoint_patch_embed.shape):
        return False

    model_shape = [model_patch_embed.shape[i] for i in range(len(model_patch_embed.shape)) if i != 1]
    checkpoint_shape = [checkpoint_patch_embed.shape[i] for i in range(len(checkpoint_patch_embed.shape)) if i != 1]
    return model_shape == checkpoint_shape

def select_patch_embed_weights(
    state_dict: dict, model: nn.Module, pretrained_bands: list[HLSBands | int | OpticalBands| SARBands], model_bands: list[HLSBands | int | OpticalBands| SARBands], proj_key: str | None = None
) -> dict:
    """Filter out the patch embedding weights according to the bands being used.
    If a band exists in the pretrained_bands, but not in model_bands, drop it.
    If a band exists in model_bands, but not pretrained_bands, randomly initialize those weights.


    Args:
        state_dict (dict): State Dict
        model (nn.Module): Model to load the weights onto.
        pretrained_bands (list[HLSBands | int]): List of bands the model was pretrained on, in the correct order.
        model_bands (list[HLSBands | int]): List of bands the model is going to be finetuned on, in the correct order
        proj_key (str, optional): Key to patch embedding projection weight in state_dict.

    Returns:
        dict: New state dict
    """
    if (type(pretrained_bands) == type(model_bands)) | (type(pretrained_bands) == int) | (type(model_bands) == int):

        if proj_key is None:
            # Search for patch embedding weight in state dict
            for key in state_dict.keys():
                if key.endswith('patch_embed.proj.weight') or key.endswith('patch_embed.projection.weight'):
                    proj_key = key
                    break
        if proj_key is None or proj_key not in state_dict:
            raise Exception("Could not find key for patch embed weight in state_dict.")

        patch_embed_weight = state_dict[proj_key]

        temp_weight = model.state_dict()[proj_key].clone()

        # only do this if the patch size and tubelet size match. If not, start with random weights
        if patch_embed_weights_are_compatible(temp_weight, patch_embed_weight):
            torch.nn.init.xavier_uniform_(temp_weight.view([temp_weight.shape[0], -1]))
            for index, band in enumerate(model_bands):
                if band in pretrained_bands:
                    logging.info(f"Loaded weights for {band} in position {index} of patch embed")
                    temp_weight[:, index] = patch_embed_weight[:, pretrained_bands.index(band)]
        else:
            warnings.warn(
                f"Incompatible shapes between patch embedding of model {temp_weight.shape} and\
                of checkpoint {patch_embed_weight.shape}",
                category=UserWarning,
                stacklevel=1,
            )

        state_dict[proj_key] = temp_weight

    return state_dict



# Copyright (c) EPFL VILAB.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# Based on timm, DPT and ConvNeXt code bases
# https://github.com/rwightman/pytorch-image-models/tree/master/timm
# https://github.com/isl-org/DPT
# https://github.com/facebookresearch/ConvNeXt
# --------------------------------------------------------

import torch
import torch.nn as nn

from .multimae_utils import DropPath


class ConvNeXtBlock(nn.Module):
    r"""ConvNeXt Block. There are two equivalent implementations:
    (1) DwConv -> LayerNorm (channels_first) -> 1x1 Conv -> GELU -> 1x1 Conv; all in (N, C, H, W)
    (2) DwConv -> Permute to (N, H, W, C); LayerNorm (channels_last) -> Linear -> GELU -> Linear; Permute back
    We use (2) as we find it slightly faster in PyTorch

    Args:
        dim (int): Number of input channels.
        drop_path: Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 0 (disabled for isotropic ConvNeXt).

    Code from: https://github.com/facebookresearch/ConvNeXt/blob/main/models/convnext.py
    """

    def __init__(self, dim, drop_path=0.0, layer_scale_init_value=0.0):
        super().__init__()
        self.dwconv = nn.Conv2d(
            dim, dim, kernel_size=7, padding=3, groups=dim
        )  # depthwise conv
        self.norm = nn.LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(
            dim, 4 * dim
        )  # pointwise/1x1 convs, implemented with linear layers
        self.act = nn.GELU()
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.gamma = (
            nn.Parameter(layer_scale_init_value * torch.ones((dim)), requires_grad=True)
            if layer_scale_init_value > 0
            else None
        )
        self.drop_path = DropPath(drop_path) if drop_path > 0.0 else nn.Identity()

    def forward(self, x):
        input = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)  # (N, C, H, W) -> (N, H, W, C)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)
        if self.gamma is not None:
            x = self.gamma * x
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)

        x = input + self.drop_path(x)
        return x


class ResidualConvUnit_custom(nn.Module):
    """Residual convolution module."""

    def __init__(self, features, activation, bn):
        """Init.
        Args:
            features (int): number of features
        """
        super().__init__()

        self.bn = bn

        self.groups = 1

        self.conv1 = nn.Conv2d(
            features,
            features,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=not self.bn,
            groups=self.groups,
        )

        self.conv2 = nn.Conv2d(
            features,
            features,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=not self.bn,
            groups=self.groups,
        )

        if self.bn == True:
            self.bn1 = nn.BatchNorm2d(features)
            self.bn2 = nn.BatchNorm2d(features)

        self.activation = activation

        self.skip_add = nn.quantized.FloatFunctional()

    def forward(self, x):
        """Forward pass.
        Args:
            x (tensor): input
        Returns:
            tensor: output
        """

        out = self.activation(x)
        out = self.conv1(out)
        if self.bn == True:
            out = self.bn1(out)

        out = self.activation(out)
        out = self.conv2(out)
        if self.bn == True:
            out = self.bn2(out)

        if self.groups > 1:
            out = self.conv_merge(out)

        return self.skip_add.add(out, x)


def make_scratch(in_shape, out_shape, groups=1, expand=False):
    scratch = nn.Module()

    out_shape1 = out_shape
    out_shape2 = out_shape
    out_shape3 = out_shape
    out_shape4 = out_shape
    if expand == True:
        out_shape1 = out_shape
        out_shape2 = out_shape * 2
        out_shape3 = out_shape * 4
        out_shape4 = out_shape * 8

    scratch.layer1_rn = nn.Conv2d(
        in_shape[0],
        out_shape1,
        kernel_size=3,
        stride=1,
        padding=1,
        bias=False,
        groups=groups,
    )
    scratch.layer2_rn = nn.Conv2d(
        in_shape[1],
        out_shape2,
        kernel_size=3,
        stride=1,
        padding=1,
        bias=False,
        groups=groups,
    )
    scratch.layer3_rn = nn.Conv2d(
        in_shape[2],
        out_shape3,
        kernel_size=3,
        stride=1,
        padding=1,
        bias=False,
        groups=groups,
    )
    scratch.layer4_rn = nn.Conv2d(
        in_shape[3],
        out_shape4,
        kernel_size=3,
        stride=1,
        padding=1,
        bias=False,
        groups=groups,
    )

    scratch.layer_rn = nn.ModuleList(
        [
            scratch.layer1_rn,
            scratch.layer2_rn,
            scratch.layer3_rn,
            scratch.layer4_rn,
        ]
    )

    return scratch


class FeatureFusionBlock_custom(nn.Module):
    """Feature fusion block."""

    def __init__(
        self,
        features,
        activation,
        deconv=False,
        bn=False,
        expand=False,
        align_corners=True,
    ):
        """Init.
        Args:
            features (int): number of features
        """
        super(FeatureFusionBlock_custom, self).__init__()

        self.deconv = deconv
        self.align_corners = align_corners

        self.groups = 1

        self.expand = expand
        out_features = features
        if self.expand == True:
            out_features = features // 2

        self.out_conv = nn.Conv2d(
            features,
            out_features,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=True,
            groups=1,
        )

        self.resConfUnit1 = ResidualConvUnit_custom(features, activation, bn)
        self.resConfUnit2 = ResidualConvUnit_custom(features, activation, bn)

        self.skip_add = nn.quantized.FloatFunctional()

    def forward(self, *xs):
        """Forward pass.
        Returns:
            tensor: output
        """
        output = xs[0]

        if len(xs) == 2:
            res = self.resConfUnit1(xs[1])
            output = self.skip_add.add(output, res)
            # output += res

        output = self.resConfUnit2(output)

        output = nn.functional.interpolate(
            output, scale_factor=2, mode="bilinear", align_corners=self.align_corners
        )

        output = self.out_conv(output)

        return output


def make_fusion_block(features, use_bn):
    return FeatureFusionBlock_custom(
        features,
        nn.ReLU(False),
        deconv=False,
        bn=use_bn,
        expand=False,
        align_corners=True,
    )


class Interpolate(nn.Module):
    """Interpolation module."""

    def __init__(self, scale_factor, mode, align_corners=False):
        """Init.
        Args:
            scale_factor (float): scaling
            mode (str): interpolation mode
        """
        super(Interpolate, self).__init__()

        self.interp = nn.functional.interpolate
        self.scale_factor = scale_factor
        self.mode = mode
        self.align_corners = align_corners

    def forward(self, x):
        """Forward pass.
        Args:
            x (tensor): input
        Returns:
            tensor: interpolated data
        """

        x = self.interp(
            x,
            scale_factor=self.scale_factor,
            mode=self.mode,
            align_corners=self.align_corners,
        )

        return x



# Copyright contributors to the Terratorch project

"""Command-line interface to TerraTorch."""

from terratorch.cli_tools import build_lightning_cli


def main():
    _ = build_lightning_cli()


if __name__ == "__main__":
    main()



import pytest

from terratorch.cli_tools import build_lightning_cli

#@pytest.mark.parametrize("suffix", ["_regression_unet", "_segmentation_unet", "_upsample_regression_unet"])
#def test_train_unet_decoder(suffix):
#    # Running the terratorch CLI
#    command_list = ["fit", "-c", f"tests/resources/configs/manufactured-train{suffix}.yaml"]
#    _ = build_lightning_cli(command_list)



from collections.abc import Sequence
from typing import Any

import albumentations as A
from torchgeo.transforms import AugmentationSequential

from terratorch.datamodules.geobench_data_module import GeobenchDataModule
from terratorch.datasets import MBrickKilnNonGeo

MEANS = {
    "COASTAL_AEROSOL": 574.7587880700896,
    "BLUE": 674.3473615470523,
    "GREEN": 886.3656479311578,
    "RED": 815.0945462528913,
    "RED_EDGE_1": 1128.8088426870465,
    "RED_EDGE_2": 1934.450471876027,
    "RED_EDGE_3": 2045.7652282437202,
    "NIR_BROAD": 2012.744587807115,
    "NIR_NARROW": 1608.6255233989034,
    "WATER_VAPOR": 1129.8171906000355,
    "CIRRUS": 83.27188605598549,
    "SWIR_1": 90.54924599052214,
    "SWIR_2": 68.98768652434848,
}

STDS = {
    "COASTAL_AEROSOL": 193.60631504991184,
    "BLUE": 238.75447480113132,
    "GREEN": 276.9631260242207,
    "RED": 361.15060137326634,
    "RED_EDGE_1": 364.5888078793488,
    "RED_EDGE_2": 724.2707123576525,
    "RED_EDGE_3": 819.653063972575,
    "NIR_BROAD": 794.3652427593881,
    "NIR_NARROW": 800.8538290702304,
    "WATER_VAPOR": 704.0219637458916,
    "CIRRUS": 36.355745901131705,
    "SWIR_1": 28.004671947623894,
    "SWIR_2": 24.268892726362033,
}


class MBrickKilnNonGeoDataModule(GeobenchDataModule):
    def __init__(
        self,
        batch_size: int = 8,
        num_workers: int = 0,
        data_root: str = "./",
        bands: Sequence[str] | None = None,
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        partition: str = "default",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            MBrickKilnNonGeo,
            MEANS,
            STDS,
            batch_size=batch_size,
            num_workers=num_workers,
            data_root=data_root,
            bands=bands,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            aug=aug,
            partition=partition,
            **kwargs,
        )



# Copyright contributors to the Terratorch project

import re
from collections.abc import Callable, Iterable

import albumentations as A
import numpy as np
import torch


def wrap_in_compose_is_list(transform_list):
    # set check shapes to false because of the multitemporal case
    return A.Compose(transform_list, is_check_shapes=False) if isinstance(transform_list, Iterable) else transform_list

def check_dataset_stackability(dataset, batch_size) -> bool:

    shapes = np.array([item["image"].shape for item in dataset])

    if np.array_equal(shapes.max(0), shapes.min(0)):
        return batch_size
    else:
        print("The batch samples can't be stacked, since they don't have the same dimensions. Setting batch_size=1.")
        return 1

class NormalizeWithTimesteps(Callable):
    def __init__(self, means, stds):
        super().__init__()
        self.means = means  # (C, T)
        self.stds = stds    # (C, T)

    def __call__(self, batch):
        image = batch["image"]

        if len(image.shape) == 5:  # (B, T, C, H, W)
            means = torch.tensor(self.means, device=image.device).transpose(0, 1).reshape(1, image.shape[1], image.shape[2], 1, 1)
            stds = torch.tensor(self.stds, device=image.device).transpose(0, 1).reshape(1, image.shape[1], image.shape[2], 1, 1)

        elif len(image.shape) == 4:  # (B, C, H, W)
            means = torch.tensor(self.means, device=image.device).mean(dim=1).view(1, image.shape[1], 1, 1)
            stds = torch.tensor(self.stds, device=image.device).mean(dim=1).view(1, image.shape[1], 1, 1)

        else:
            msg = f"Expected batch to have 5 or 4 dimensions, but got {len(image.shape)}"
            raise Exception(msg)

        batch["image"] = (image - means) / stds
        return batch



import glob
import os
from pathlib import Path
from typing import Any

import albumentations as A
import geopandas
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rioxarray
import torch
from matplotlib import cm
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from torch import Tensor
from torchgeo.datasets import NonGeoDataset

from terratorch.datasets.utils import default_transform, filter_valid_files


class Sen1Floods11NonGeo(NonGeoDataset):
    """NonGeo dataset implementation for sen1floods11."""
    all_band_names = (
                "COASTAL_AEROSOL",
                "BLUE",
                "GREEN",
                "RED",
                "RED_EDGE_1",
                "RED_EDGE_2",
                "RED_EDGE_3",
                "NIR_BROAD",
                "NIR_NARROW",
                "WATER_VAPOR",
                "CIRRUS",
                "SWIR_1",
                "SWIR_2",
            )

    rgb_bands = ("RED", "GREEN", "BLUE")
    BAND_SETS = {"all": all_band_names, "rgb": rgb_bands}
    num_classes = 2
    splits = {"train": "train", "val": "valid", "test": "test"}
    data_dir = "v1.1/data/flood_events/HandLabeled/S2Hand"
    label_dir = "v1.1/data/flood_events/HandLabeled/LabelHand"
    split_dir = "v1.1/splits/flood_handlabeled"
    metadata_file = "v1.1/Sen1Floods11_Metadata.geojson"

    def __init__(
        self,
        data_root: str,
        split="train",
        bands: None | list[int] = None,
        transform: A.Compose | None = None,
    ) -> None:
        super().__init__()
        if split not in ["train", "test", "val"]:
            msg = "Split must be one of train, test, val."
            raise Exception(msg)
        if split == "val":
            split = "valid"

        self.bands = bands
        data_root = Path(data_root)
        data_directory = data_root / "data/data/flood_events/HandLabeled/"
        input_directory = data_directory / "S2Hand"
        label_directory = data_directory / "LabelHand"

        # split_file = data_root / f"splits/splits/flood_handlabeled/flood_bolivia_data_S2.txt"
        split_file = data_root / f"splits/splits/flood_handlabeled/flood_{split}_data_S2_geodn.txt"
        # split_file = data_root / f"splits/splits/flood_handlabeled/flood_bolivia_data_S2.txt"
        metadata_file = data_root / "Sen1Floods11_Metadata.geojson"
        self.metadata = geopandas.read_file(metadata_file)

        self.image_files = sorted(glob.glob(os.path.join(input_directory, "*.tif")))
        self.segmentation_mask_files = sorted(glob.glob(os.path.join(label_directory, "*.tif")))

        with open(split_file) as f:
            split = f.readlines()
        valid_files = {rf"{substring.strip()}" for substring in split}
        self.image_files = filter_valid_files(
            self.image_files,
            valid_files=valid_files,
            ignore_extensions=True,
            allow_substring=True,
        )
        self.segmentation_mask_files = filter_valid_files(
            self.segmentation_mask_files,
            valid_files=valid_files,
            ignore_extensions=True,
            allow_substring=True,
        )

        self.rgb_indices = [2, 1, 0]
        self.transform = transform if transform else default_transform

    def __len__(self) -> int:
        return len(self.image_files)

    def _get_date(self, index) -> np.ndarray:
        # move this logic to the model?
        file_name = self.image_files[index]
        location = os.path.basename(file_name).split("_")[0]
        if self.metadata[self.metadata["location"] == location].shape[0] != 1:
            # msg = f"No date found for sample {file_name}"
            # raise Exception(msg)
            date = pd.to_datetime("13-10-1998", dayfirst=True)
        else:
            date = pd.to_datetime(self.metadata[self.metadata["location"] == location]["s2_date"].item())
        date_np = np.zeros((1, 3))
        date_np[0, 0] = date.year
        date_np[0, 1] = date.dayofyear - 1  # base 0
        # date_np[0, 2] = date.hour
        return date_np

    def _get_coords(self, index) -> np.ndarray:
        file_name = self.image_files[index]
        image = rioxarray.open_rasterio(file_name)
        # lons_lats = np.meshgrid(image.x / 180, image.y / 90)
        # coords = np.stack([np.stack(lon_lat) for lon_lat in lons_lats])
        # coords shape: batch_size, 2 (lon, lat), height, width

        lat_lon = np.array([image.y[image.shape[0]//2], image.x[image.shape[1]//2]])
        return lat_lon

    def __getitem__(self, index: int) -> dict[str, Any]:
        image = self._load_file(self.image_files[index]).astype(np.float32) * 0.0001
        if self.bands:
            image = image[self.bands, ...]
        image = np.moveaxis(image, 0, -1)
        output = {
            "image": image,
            "mask": self._load_file(self.segmentation_mask_files[index]).astype(np.int64)[0],
        }
        if self.transform:
            output = self.transform(**output)

        output["location_coords"] = np.moveaxis(self._get_coords(index).astype(np.float32), 0, -1)
        output["temporal_coords"] = self._get_date(index).astype(np.float32)
        return output

    def _load_file(self, path: Path):
        data = rioxarray.open_rasterio(path)
        return data.to_numpy()

    def plot(self, sample: dict[str, Tensor], suptitle: str | None = None) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        image = sample["image"]
        if torch.is_tensor(image):
            image = image.numpy()
        image = image.take(self.rgb_indices, axis=0)
        image = np.transpose(image, (1, 2, 0))
        image = (image - image.min(axis=(0, 1))) * (1 / image.max(axis=(0, 1)))
        image = np.clip(image, 0, 1)

        label_mask = sample["mask"]
        label_mask = np.transpose(label_mask, (1, 2, 0))

        showing_predictions = "prediction" in sample
        if showing_predictions:
            prediction_mask = sample["prediction"]

        return self._plot_sample(
            image,
            label_mask,
            prediction=prediction_mask if showing_predictions else None,
            suptitle=suptitle,
        )

    @staticmethod
    def _plot_sample(image, label, num_classes, prediction=None, suptitle=None, class_names=None):
        num_images = 5 if prediction else 4
        fig, ax = plt.subplots(1, num_images, figsize=(8, 6))

        # for legend
        ax[0].axis("off")

        norm = mpl.colors.Normalize(vmin=0, vmax=num_classes - 1)
        ax[1].axis("off")
        ax[1].title.set_text("Image")
        ax[1].imshow(image)

        ax[2].axis("off")
        ax[2].title.set_text("Ground Truth Mask")
        ax[2].imshow(label, cmap="jet", norm=norm)

        ax[3].axis("off")
        ax[3].title.set_text("GT Mask on Image")
        ax[3].imshow(image)
        ax[3].imshow(label, cmap="jet", alpha=0.3, norm=norm)

        if prediction:
            ax[4].title.set_text("Predicted Mask")
            ax[4].imshow(prediction, cmap="jet", norm=norm)

        cmap = cm.get_cmap("jet")
        legend_data = []
        for i, _ in enumerate(range(num_classes)):
            class_name = class_names[i] if class_names else str(i)
            data = [i, cmap(norm(i)), class_name]
            legend_data.append(data)
        handles = [Rectangle((0, 0), 1, 1, color=tuple(v for v in c)) for k, c, n in legend_data]
        labels = [n for k, c, n in legend_data]
        ax[0].legend(handles, labels, loc="center")
        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



# Copyright contributors to the Terratorch project

from random import choice
from typing import Iterable, Iterator

from torch.utils.data import Sampler
from torchgeo.datasets import BoundingBox
from torchgeo.samplers import GeoSampler


class MultiSampler(Sampler[BoundingBox]):
    def __init__(self, samplers: Iterable[GeoSampler]) -> None:
        self.length = sum(map(len, samplers))
        self.samplers = samplers
    
    def __len__(self) -> int:
        return self.length


class RandomMultiSampler(MultiSampler):
    def __iter__(self) -> Iterator[BoundingBox]:
        sampler_set = set(list(iter(sampler) for sampler in self.samplers))
        sampler_list = list(sampler_set)
        while len(sampler_set) > 0:
            sampler = choice(sampler_list)
            try:
                yield next(sampler)
            except StopIteration:
                sampler_set.remove(sampler)
                sampler_list = list(sampler_set)


class SequentialMultiSampler(MultiSampler):
    def __iter__(self) -> Iterator[BoundingBox]:
        sampler_iterables = list(iter(sampler) for sampler in self.samplers)
        for sampler in sampler_iterables:
            for sample in sampler:
                yield sample



"""This module contains logic for tiled inference.
    This does some additional things over the obvious fold -> predict -> unfold logic,
    e.g. cropping out areas around model prediction to reduce artifacts

    It additionally rebatches after the fold operation to gain speed up.
"""

from collections.abc import Callable
from dataclasses import dataclass

import torch


@dataclass
class TiledInferenceParameters:
    """Parameters to be used for inference.

    Args:
        h_crop (int): height of the crop region
        h_stride (int): size of the stride on the y-axis
        w_crop (int): width of the crop region
        w_stride (int): size of the stride on the x-axis
        delta (int): size of the border cropped from each tile. Defaults to None, which computes this automatically,
          with a minimum of 16.
        average_patches (bool): Whether to average the overlapping regions. Defaults to True.
    """

    h_crop: int
    h_stride: int
    w_crop: int
    w_stride: int
    delta: int = None
    average_patches: bool = True


@dataclass
class InferenceInput:
    batch: int
    input_coords: tuple[slice, slice]
    input_data: torch.Tensor
    output_crop: None | tuple[slice, slice]


def tiled_inference(
    model_forward: Callable,
    input_batch: torch.Tensor,
    out_channels: int,
    inference_parameters: TiledInferenceParameters,
) -> torch.Tensor:
    """
    Like divide an image into (potentially) overlapping tiles and perform inference on them.
    Additionally rebatch for increased GPU utilization.

    Args:
        model_forward (Callable): Callable that return the output of the model.
        input_batch (torch.Tensor): Input batch to be processed
        out_channels (int): Number of output channels
        inference_parameters (TiledInferenceParameters): Parameters to be used for the process.

    Returns:
        torch.Tensor: The result of the inference
    """

    batch_size, h_img, w_img = input_batch.shape[0], *input_batch.shape[-2:]
    preds = input_batch.new_zeros((batch_size, out_channels, h_img, w_img))

    # this list will contain tuples. Inside the tuples:
    #   0. batch
    #   1. Coordinates where this should end up in the preds
    #   2. output/input
    #   3. Optionally, for inputs, how to crop the output
    # it is important this allocation follows the same order as in the original code

    # Stage 1: deal with border patches
    # Deal with patches near the right border

    coordinates_and_inputs: list[InferenceInput] = []
    for i in range(0, h_img - inference_parameters.h_crop + 1, inference_parameters.h_crop):
        patch = input_batch[..., i : i + inference_parameters.h_crop, w_img - inference_parameters.w_crop : w_img]
        coordinates_and_inputs += [
            InferenceInput(
                b,
                (slice(i, i + inference_parameters.h_crop), slice(w_img - inference_parameters.w_crop, w_img)),
                patch[b],
                None,
            )
            for b in range(batch_size)
        ]

    # Deal with patches near the bottom of the image
    for i in range(0, w_img - inference_parameters.w_crop + 1, inference_parameters.w_crop):
        patch = input_batch[..., h_img - inference_parameters.h_crop : h_img, i : i + inference_parameters.w_crop]
        coordinates_and_inputs += [
            InferenceInput(
                b,
                (slice(h_img - inference_parameters.h_crop, h_img), slice(i, i + inference_parameters.w_crop)),
                patch[b],
                None,
            )
            for b in range(batch_size)
        ]

    # Deal with last patches at the right bottom of the image
    patch = input_batch[..., h_img - inference_parameters.h_crop : h_img, w_img - inference_parameters.w_crop : w_img]
    coordinates_and_inputs += [
        InferenceInput(
            b,
            (slice(h_img - inference_parameters.h_crop, h_img), slice(w_img - inference_parameters.w_crop, w_img)),
            patch[b],
            None,
        )
        for b in range(batch_size)
    ]

    if inference_parameters.delta:
        delta_x = inference_parameters.delta
        delta_y = inference_parameters.delta
    else:
        delta_x = min(16, (inference_parameters.w_crop - inference_parameters.w_stride) // 2)
        delta_y = min(16, (inference_parameters.h_crop - inference_parameters.h_stride) // 2)

    # Stage 2: process internally with patch overlap {2*delta/inference_parameters.w_crop}
    nr = h_img
    nc = w_img
    for row in range(0, nr - inference_parameters.h_crop + 1, inference_parameters.h_stride):
        for col in range(0, nc - inference_parameters.w_crop + 1, inference_parameters.w_stride):
            patch = input_batch[..., row : row + inference_parameters.h_crop, col : col + inference_parameters.w_crop]
            if row == 0 or col == 0:
                coordinates_and_inputs += [
                    InferenceInput(
                        b,
                        (slice(row, row + inference_parameters.h_crop), slice(col, col + inference_parameters.w_crop)),
                        patch[b],
                        None,
                    )
                    for b in range(batch_size)
                ]
            else:
                coordinates_and_inputs += [
                    InferenceInput(
                        b,
                        (
                            slice(row + delta_y, row + inference_parameters.h_crop - delta_y),
                            slice(col + delta_x, col + inference_parameters.w_crop - delta_x),
                        ),
                        patch[b],
                        (
                            slice(delta_y, inference_parameters.h_crop - delta_y),
                            slice(delta_x, inference_parameters.w_crop - delta_x),
                        ),
                    )
                    for b in range(batch_size)
                ]

    # NOTE: the output may be SLIGHTLY different using batched inputs because of layers such as nn.LayerNorm
    # During inference, these layers compute batch statistics that affect the output.
    # However, this should still be correct.
    # TODO: make this configurable by user?
    process_batch_size = 16
    with torch.no_grad():
        preds_count = input_batch.new_zeros(batch_size, preds.shape[-2], preds.shape[-1])
        for start in range(0, len(coordinates_and_inputs), process_batch_size):
            end = min(len(coordinates_and_inputs), start + process_batch_size)
            batch = coordinates_and_inputs[start:end]
            tensor_input = torch.stack([b.input_data for b in batch], dim=0)
            output = model_forward(tensor_input)
            output = [output[i] for i in range(len(batch))]
            for batch_input, predicted in zip(batch, output, strict=True):
                if batch_input.output_crop is not None:
                    predicted = predicted[..., batch_input.output_crop[0], batch_input.output_crop[1]]
                if inference_parameters.average_patches:
                    preds[
                        batch_input.batch,
                        :,
                        batch_input.input_coords[0],
                        batch_input.input_coords[1],
                    ] += predicted
                else:
                    preds[
                        batch_input.batch,
                        :,
                        batch_input.input_coords[0],
                        batch_input.input_coords[1],
                    ] = predicted

                preds_count[
                    batch_input.batch,
                    batch_input.input_coords[0],
                    batch_input.input_coords[1],
                ] += 1

    if (preds_count == 0).sum() != 0:
        msg = "Some pixels did not receive a classification!"
        raise RuntimeError(msg)
    if inference_parameters.average_patches:
        return preds / preds_count.unsqueeze(1)
    return preds



# Copyright (c) EPFL VILAB.
# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
# --------------------------------------------------------
# Based on timm, DeiT, DINO, MoCo-v3, BEiT, MAE-priv and MAE code bases
# https://github.com/rwightman/pytorch-image-models/tree/master/timm
# https://github.com/facebookresearch/deit
# https://github.com/facebookresearch/dino
# https://github.com/facebookresearch/moco-v3
# https://github.com/microsoft/unilm/tree/master/beit
# https://github.com/BUPT-PRIV/MAE-priv
# https://github.com/facebookresearch/mae
# --------------------------------------------------------

import torch
import torch.nn as nn
import torch.nn.functional as F
from einops import rearrange


class MaskedCrossEntropyLoss(nn.Module):
    """Cross-entropy loss with masking
    :param patch_size: Patch size
    :param stride: Stride of task / modality
    :param label_smoothing: Amount of smoothing in the loss (default is 0.0)
    """

    def __init__(
        self, patch_size: int = 16, stride: int = 1, label_smoothing: float = 0.0
    ):
        super().__init__()
        self.patch_size = patch_size
        self.stride = stride
        self.scale_factor = patch_size // stride
        self.label_smoothing = label_smoothing

    def forward(self, input, target, mask=None):

        loss = F.cross_entropy(
            input, target, reduction="none", label_smoothing=self.label_smoothing
        )

        if mask is not None:
            if mask.sum() == 0:
                return torch.tensor(0).to(loss.device)

            H, W = input.shape[-2:]
            nh, nw = H // self.scale_factor, W // self.scale_factor
            # Resize mask and upsample
            mask = rearrange(mask, "b (nh nw) -> b nh nw", nh=nh, nw=nw)
            mask = F.interpolate(
                mask.unsqueeze(1).float(), size=(H, W), mode="nearest"
            ).squeeze(1)
            loss = loss * mask
            # Compute mean per sample
            loss = loss.flatten(start_dim=1).sum(dim=1) / mask.flatten(start_dim=1).sum(
                dim=1
            )
            loss = loss.nanmean()  # Account for zero masks
        else:
            loss = loss.mean()  # If this is ever nan, we want it to stop training

        return loss


class MaskedMSELoss(nn.Module):
    """L1 loss with masking
    :param patch_size: Patch size
    :param stride: Stride of task / modality
    :param norm_pix: Normalized pixel loss
    """

    def __init__(self, patch_size: int = 16, stride: int = 1, norm_pix=False):
        super().__init__()
        self.patch_size = patch_size
        self.stride = stride
        self.scale_factor = patch_size // stride
        self.norm_pix = norm_pix

    def patchify(self, imgs, nh, nw):
        p = self.scale_factor
        x = rearrange(
            imgs, "b c (nh p1) (nw p2) -> b (nh nw) (p1 p2 c)", nh=nh, nw=nw, p1=p, p2=p
        )
        return x

    def unpatchify(self, x, nh, nw):
        p = self.scale_factor
        imgs = rearrange(
            x, "b (nh nw) (p1 p2 c) -> b c (nh p1) (nw p2)", nh=nh, nw=nw, p1=p, p2=p
        )
        return imgs

    def forward(self, input, target, mask=None):

        H, W = input.shape[-2:]
        nh, nw = H // self.scale_factor, W // self.scale_factor

        if self.norm_pix:
            target = self.patchify(target, nh, nw)
            mean = target.mean(dim=-1, keepdim=True)
            var = target.var(dim=-1, keepdim=True)
            eps = 1e-6
            target = (target - mean) / torch.sqrt(var + eps)
            target = self.unpatchify(target, nh, nw)

        loss = F.mse_loss(input, target, reduction="none")

        if mask is not None:
            if mask.sum() == 0:
                return torch.tensor(0).to(loss.device)

            # Resize mask and upsample
            mask = rearrange(mask, "b (nh nw) -> b nh nw", nh=nh, nw=nw)
            mask = F.interpolate(
                mask.unsqueeze(1).float(), size=(H, W), mode="nearest"
            ).squeeze(1)
            loss = loss.mean(dim=1)  # B, C, H, W -> B, H, W
            loss = loss * mask
            # Compute mean per sample
            loss = loss.flatten(start_dim=1).sum(dim=1) / mask.flatten(start_dim=1).sum(
                dim=1
            )
            loss = loss.nanmean()  # Account for zero masks
        else:
            loss = loss.mean()  # If this is ever nan, we want it to stop training

        return loss


class MaskedL1Loss(nn.Module):
    """L1 loss with masking
    :param patch_size: Patch size
    :param stride: Stride of task / modality
    :param norm_pix: Normalized pixel loss
    """

    def __init__(self, patch_size: int = 16, stride: int = 1, norm_pix=False):
        super().__init__()
        self.patch_size = patch_size
        self.stride = stride
        self.scale_factor = patch_size // stride
        self.norm_pix = norm_pix

    def patchify(self, imgs, nh, nw):
        p = self.scale_factor
        x = rearrange(
            imgs, "b c (nh p1) (nw p2) -> b (nh nw) (p1 p2 c)", nh=nh, nw=nw, p1=p, p2=p
        )
        return x

    def unpatchify(self, x, nh, nw):
        p = self.scale_factor
        imgs = rearrange(
            x, "b (nh nw) (p1 p2 c) -> b c (nh p1) (nw p2)", nh=nh, nw=nw, p1=p, p2=p
        )
        return imgs

    def forward(self, input, target, mask=None):

        H, W = input.shape[-2:]
        nh, nw = H // self.scale_factor, W // self.scale_factor

        if self.norm_pix:
            target = self.patchify(target, nh, nw)
            mean = target.mean(dim=-1, keepdim=True)
            var = target.var(dim=-1, keepdim=True)
            eps = 1e-6
            target = (target - mean) / torch.sqrt(var + eps)
            target = self.unpatchify(target, nh, nw)

        loss = F.l1_loss(input, target, reduction="none")

        if mask is not None:
            if mask.sum() == 0:
                return torch.tensor(0).to(loss.device)

            # Resize mask and upsample
            mask = rearrange(mask, "b (nh nw) -> b nh nw", nh=nh, nw=nw)
            mask = F.interpolate(
                mask.unsqueeze(1).float(), size=(H, W), mode="nearest"
            ).squeeze(1)
            loss = loss.mean(dim=1)  # B, C, H, W -> B, H, W
            loss = loss * mask
            # Compute mean per sample
            loss = loss.flatten(start_dim=1).sum(dim=1) / mask.flatten(start_dim=1).sum(
                dim=1
            )
            loss = loss.nanmean()  # Account for zero masks
        else:
            loss = loss.mean()  # If this is ever nan, we want it to stop training

        return loss



import os
import shutil
import gc
import pytest
import timm
import torch

from terratorch.cli_tools import build_lightning_cli
from terratorch.registry import BACKBONE_REGISTRY

@pytest.fixture(autouse=True)
def setup_and_cleanup(model_name):
    model_instance = BACKBONE_REGISTRY.build(model_name)

    state_dict = model_instance.state_dict()

    torch.save(state_dict, os.path.join("tests", model_name + ".pt"))

    yield # everything after this runs after each test

    os.remove(os.path.join("tests", model_name + ".pt"))

    if os.path.isdir(os.path.join("tests", "all_ecos_random")):
        shutil.rmtree(os.path.join("tests", "all_ecos_random"))

@pytest.mark.parametrize("model_name", ["prithvi_eo_v1_100", "prithvi_eo_v2_300", "prithvi_swin_B", "prithvi_swin_L", "prithvi_eo_v2_600"])
@pytest.mark.parametrize("case", ["fit", "test", "validate"])
def test_finetune_multiple_backbones(model_name, case):
    command_list = [case, "-c", f"tests/resources/configs/manufactured-finetune_{model_name}.yaml"]
    _ = build_lightning_cli(command_list)

    gc.collect()

@pytest.mark.parametrize("model_name", ["prithvi_swin_B"])
@pytest.mark.parametrize("case", ["fit", "test", "validate"])
def test_finetune_bands_intervals(model_name, case):
    command_list = [case, "-c", f"tests/resources/configs/manufactured-finetune_{model_name}_band_interval.yaml"]
    _ = build_lightning_cli(command_list)

    gc.collect()

@pytest.mark.parametrize("model_name", ["prithvi_swin_B"])
@pytest.mark.parametrize("case", ["fit", "test", "validate"])
def test_finetune_bands_str(model_name, case):
    command_list = [case, "-c", f"tests/resources/configs/manufactured-finetune_{model_name}_string.yaml"]
    _ = build_lightning_cli(command_list)

    gc.collect()

@pytest.mark.parametrize("model_name", ["prithvi_eo_v2_300"])
@pytest.mark.parametrize("case", ["fit", "test", "validate"])
def test_finetune_pad(case):
    command_list = [case, "-c", f"tests/resources/configs/manufactured-finetune_prithvi_pixelwise_pad.yaml"]
    _ = build_lightning_cli(command_list)

    gc.collect()

@pytest.mark.parametrize("model_name", ["prithvi_swin_B"])
def test_finetune_metrics_from_file(model_name):

    model_instance = timm.create_model(model_name)

    state_dict = model_instance.state_dict()

    torch.save(state_dict, os.path.join("tests/resources/configs/", model_name + ".pt"))

    # Running the terratorch CLI
    command_list = ["fit", "-c", f"tests/resources/configs/manufactured-finetune_{model_name}_metrics_from_file.yaml"]
    _ = build_lightning_cli(command_list)

    gc.collect()




# Copyright contributors to the Terratorch project

"""
This module contains generic data modules for instantiation at runtime.
"""
import os
import logging
from collections.abc import Callable, Iterable
from pathlib import Path
from typing import Any, Iterator
import albumentations as A
import numpy as np
import torch
from torch import Tensor
from torch.utils.data import DataLoader, RandomSampler, BatchSampler, SequentialSampler, default_collate
from torchgeo.datamodules import NonGeoDataModule

from terratorch.datasets import (GenericMultimodalDataset, GenericMultimodalSegmentationDataset,
                                 GenericMultimodalPixelwiseRegressionDataset, GenericMultimodalScalarDataset, HLSBands)
from terratorch.datamodules.generic_pixel_wise_data_module import Normalize
from terratorch.io.file import load_from_file_or_attribute

logger = logging.getLogger("terratorch")

def collate_chunk_dicts(batch_list):
    batch = {}
    for key, value in batch_list[0].items():  # TODO: Handle missing modalities when allow_missing_modalities is set.
        if isinstance(value, torch.Tensor):
            batch[key] = torch.concat([chunk[key] for chunk in batch_list])
        elif isinstance(value, np.ndarray):
            batch[key] = np.concatenate([chunk[key] for chunk in batch_list])
        elif isinstance(value, dict):
            batch[key] = collate_chunk_dicts([chunk[key] for chunk in batch_list])
        else:
            batch[key] = [chunk[key] for chunk in batch_list]
    return batch


def collate_samples(batch_list):
    """
    Wrapper for default_collate as it cannot handle some datatypes such as np.datetime64.
    """
    batch = {}
    for key, value in batch_list[0].items():  # TODO: Handle missing modalities when allow_missing_modalities is set.
        if isinstance(value, dict):
            batch[key] = collate_samples([chunk[key] for chunk in batch_list])
        else:
            try:
                batch[key] = default_collate([chunk[key] for chunk in batch_list])
            except TypeError:
                # Fallback to numpy or simple list
                if isinstance(value, np.ndarray):
                    batch[key] = np.stack([chunk[key] for chunk in batch_list])
                else:
                    batch[key] = [chunk[key] for chunk in batch_list]
    return batch


def wrap_in_compose_is_list(transform_list, image_modalities=None, non_image_modalities=None):
    additional_targets = {}
    if image_modalities:
        for modality in image_modalities:
            additional_targets[modality] = "image"
    if non_image_modalities:
        # Global label values are ignored and need to be processed separately
        for modality in non_image_modalities:
            additional_targets[modality] = "global_label"
    # set check shapes to false because of the multitemporal case
    return A.Compose(transform_list, is_check_shapes=False, additional_targets=additional_targets) \
        if isinstance(transform_list, Iterable) else transform_list

class MultimodalNormalize(Callable):
    def __init__(self, means, stds):
        super().__init__()
        self.means = means
        self.stds = stds

    def __call__(self, batch):
        for m in self.means.keys():
            if m not in batch["image"]:
                continue
            image = batch["image"][m]
            if len(image.shape) == 5:
                # B, C, T, H, W
                means = torch.tensor(self.means[m], device=image.device).view(1, -1, 1, 1, 1)
                stds = torch.tensor(self.stds[m], device=image.device).view(1, -1, 1, 1, 1)
            elif len(image.shape) == 4:
                # B, C, H, W
                means = torch.tensor(self.means[m], device=image.device).view(1, -1, 1, 1)
                stds = torch.tensor(self.stds[m], device=image.device).view(1, -1, 1, 1)
            elif len(self.means[m]) == 1:
                # B, (T,) H, W
                means = torch.tensor(self.means[m], device=image.device)
                stds = torch.tensor(self.stds[m], device=image.device)
            elif len(image.shape) == 3:  # No batch dim
                # C, H, W
                means = torch.tensor(self.means[m], device=image.device).view(-1, 1, 1)
                stds = torch.tensor(self.stds[m], device=image.device).view(-1, 1, 1)

            elif len(image.shape) == 2:
                means = torch.tensor(self.means[m], device=image.device)
                stds = torch.tensor(self.stds[m], device=image.device)

            elif len(image.shape) == 1:
                means = torch.tensor(self.means[m], device=image.device)
                stds = torch.tensor(self.stds[m], device=image.device)

            else:
                msg = (f"Expected batch with 5 or 4 dimensions (B, C, (T,) H, W), sample with 3 dimensions (C, H, W) "
                       f"or a single channel, but got {len(image.shape)}")
                raise Exception(msg)
            batch["image"][m] = (image - means) / stds
        return batch


class MultiModalBatchSampler(BatchSampler):
    """
    Sample a defined number of modalities per batch (see sample_num_modalities and sample_replace)
    """
    def __init__(self, modalities, sample_num_modalities, sample_replace, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.modalities = modalities
        self.sample_num_modalities = sample_num_modalities
        self.sample_replace = sample_replace

    def __iter__(self) -> Iterator[list[int]]:
        """
        Code similar to BatchSampler but samples tuples in the format (idx, ["m1", "m2", ...])
        """
        # Select sampled modalities per batch
        sampled_modalities = np.random.choice(self.modalities, self.sample_num_modalities, replace=self.sample_replace)

        if self.drop_last:
            sampler_iter = iter(self.sampler)
            while True:
                try:
                    batch = [(next(sampler_iter), sampled_modalities) for _ in range(self.batch_size)]
                    yield batch
                except StopIteration:
                    break
        else:
            batch = [0] * self.batch_size
            idx_in_batch = 0
            for idx in self.sampler:
                batch[idx_in_batch] = (idx, sampled_modalities)
                idx_in_batch += 1
                if idx_in_batch == self.batch_size:
                    yield batch
                    idx_in_batch = 0
                    batch = [0] * self.batch_size
            if idx_in_batch > 0:
                yield batch[:idx_in_batch]


class GenericMultiModalDataModule(NonGeoDataModule):
    """
    This is a generic datamodule class for instantiating data modules at runtime.
    Composes several [GenericNonGeoSegmentationDatasets][terratorch.datasets.GenericNonGeoSegmentationDataset]
    """

    def __init__(
        self,
        batch_size: int,
        modalities: list[str],
        train_data_root: dict[str, Path],
        val_data_root: dict[str, Path],
        test_data_root: dict[str, Path],
        means: dict[str, list],
        stds: dict[str, list],
        task: str | None = None,
        num_classes: int | None = None,
        image_grep: str | dict[str, str] | None = None,
        label_grep: str | None = None,
        train_label_data_root: Path | str | None = None,
        val_label_data_root: Path | str | None = None,
        test_label_data_root: Path | str | None = None,
        predict_data_root: dict[str, Path] | str | None = None,
        train_split: Path | str | None = None,
        val_split: Path | str | None = None,
        test_split: Path| str | None = None,
        dataset_bands: dict[str, list] | None = None,
        output_bands: dict[str, list] | None = None,
        predict_dataset_bands: dict[str, list] | None = None,
        predict_output_bands: dict[str, list] | None = None,
        image_modalities: list[str] | None = None,
        rgb_modality: str | None = None,
        rgb_indices: list[int] | None = None,
        allow_substring_file_names: bool = False,
        class_names: list[str] | None = None,
        constant_scale: dict[float] = None,
        train_transform: dict | A.Compose | None | list[A.BasicTransform] = None,
        val_transform: dict | A.Compose | None | list[A.BasicTransform] = None,
        test_transform: dict | A.Compose | None | list[A.BasicTransform] = None,
        shared_transforms: list | bool = True,
        expand_temporal_dimension: bool = False,
        no_data_replace: float | None = None,
        no_label_replace: float | None = -1,
        reduce_zero_label: bool = False,
        drop_last: bool = True,
        num_workers: int = 0,
        pin_memory: bool = False,
        data_with_sample_dim: bool = False,
        sample_num_modalities: int | None = None,
        sample_replace: bool = False,
        channel_position: int = -3,
        concat_bands: bool = False,
        **kwargs: Any,
    ) -> None:
        """Constructor

        Args:
            batch_size (int): Number of samples in per batch.
            modalities (list[str]): List of modalities.
            train_data_root (dict[Path]): Dictionary of paths to training data root directory or csv/parquet files with 
                image-level data, with modalities as keys.
            val_data_root (dict[Path]): Dictionary of paths to validation data root directory or csv/parquet files with 
                image-level data, with modalities as keys.
            test_data_root (dict[Path]): Dictionary of paths to test data root directory or csv/parquet files with 
                image-level data, with modalities as keys.
            means (dict[list]): Dictionary of mean values as lists with modalities as keys.
            stds (dict[list]): Dictionary of std values as lists with modalities as keys.
            task (str, optional): Selected task form segmentation, regression (pixel-wise), classification,
                multilabel_classification, scalar_regression, scalar (custom image-level task), or None (no targets).
                Defaults to None.
            num_classes (int, optional): Number of classes in classification or segmentation tasks.
            predict_data_root (dict[Path], optional): Dictionary of paths to data root directory or csv/parquet files
                with image-level data, with modalities as keys.
            image_grep (dict[str], optional): Dictionary with regular expression appended to data_root to find input
                images, with modalities as keys. Defaults to "*". Ignored when allow_substring_file_names is False.
            label_grep (str, optional): Regular expression appended to label_data_root to find labels or mask files.
                Defaults to "*". Ignored when allow_substring_file_names is False.
            train_label_data_root (Path | None, optional): Path to data root directory with training labels or
                csv/parquet files with labels. Required for supervised tasks.
            val_label_data_root (Path | None, optional): Path to data root directory with validation labels or
                csv/parquet files with labels. Required for supervised tasks.
            test_label_data_root (Path | None, optional): Path to data root directory with test labels or
                csv/parquet files with labels. Required for supervised tasks.
            train_split (Path, optional): Path to file containing training samples prefixes to be used for this split.
                The file can be a csv/parquet file with the prefixes in the index or a txt file with new-line separated
                sample prefixes. File names must be exact matches if allow_substring_file_names is False. Otherwise,
                files are searched using glob with the form Path(data_root).glob(prefix + [image or label grep]).
                If not specified, search samples based on files in data_root. Defaults to None.
            val_split (Path, optional): Path to file containing validation samples prefixes to be used for this split.
                The file can be a csv/parquet file with the prefixes in the index or a txt file with new-line separated
                sample prefixes. File names must be exact matches if allow_substring_file_names is False. Otherwise,
                files are searched using glob with the form Path(data_root).glob(prefix + [image or label grep]).
                If not specified, search samples based on files in data_root. Defaults to None.
            test_split (Path, optional): Path to file containing test samples prefixes to be used for this split.
                The file can be a csv/parquet file with the prefixes in the index or a txt file with new-line separated
                sample prefixes. File names must be exact matches if allow_substring_file_names is False. Otherwise,
                files are searched using glob with the form Path(data_root).glob(prefix + [image or label grep]).
                If not specified, search samples based on files in data_root. Defaults to None.
            dataset_bands (dict[list], optional): Bands present in the dataset, provided in a dictionary with modalities
                as keys. This parameter names input channels (bands) using HLSBands, ints, int ranges, or strings, so
                that they can then be referred to by output_bands. Needs to be superset of output_bands. Can be a subset
                of all modalities. Defaults to None.
            output_bands (dict[list], optional): Bands that should be output by the dataset as named by dataset_bands,
                provided as a dictionary with modality keys. Can be subset of all modalities. Defaults to None.
            predict_dataset_bands (list[dict], optional): Overwrites dataset_bands with this value at predict time.
                Defaults to None, which does not overwrite.
            predict_output_bands (list[dict], optional): Overwrites output_bands with this value at predict time.
                Defaults to None, which does not overwrite.
            image_modalities(list[str], optional): List of pixel-level raster modalities. Defaults to data_root.keys().
                The difference between all modalities and image_modalities are non-image modalities which are treated
                differently during the transforms and are not modified but only converted into a tensor if possible.
            rgb_modality (str, optional): Modality used for RGB plots. Defaults to first modality in data_root.keys().
            rgb_indices (list[int] | None, optional): _description_. Defaults to None.
            allow_substring_file_names
            class_names (list[str], optional): Names of the classes. Defaults to None.
            constant_scale (dict[float]): Factor to multiply data values by, provided as a dictionary with modalities as
                keys. Can be subset of all modalities. Defaults to None.
            train_transform (Albumentations.Compose | dict | None): Albumentations transform to be applied to all image
                modalities. Should end with ToTensorV2() and not include normalization. The transform is not applied to 
                non-image data, which is only converted to tensors if possible. If dict, can include separate transforms 
                per modality (no shared parameters between modalities). 
                Defaults to None, which simply applies ToTensorV2().
            val_transform (Albumentations.Compose | dict | None): Albumentations transform to be applied to all image
                modalities. Should end with ToTensorV2() and not include normalization. The transform is not applied to 
                non-image data, which is only converted to tensors if possible. If dict, can include separate transforms 
                per modality (no shared parameters between modalities). 
                Defaults to None, which simply applies ToTensorV2().
            test_transform (Albumentations.Compose | dict | None): Albumentations transform to be applied to all image
                modalities. Should end with ToTensorV2() and not include normalization. The transform is not applied to 
                non-image data, which is only converted to tensors if possible. If dict, can include separate transforms 
                per modality (no shared parameters between modalities). 
                Defaults to None, which simply applies ToTensorV2().
            shared_transforms (bool): transforms are shared between all image modalities (e.g., similar crop). 
                This setting is ignored if transforms are defined per modality. Defaults to True.  
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Only works with image modalities. Is only applied to modalities with defined dataset_bands.
                Defaults to False.
            no_data_replace (float | None): Replace nan values in input images with this value. If none, does no 
                replacement. Defaults to None.
            no_label_replace (int | None): Replace nan values in label with this value. If none, does no replacement. 
                Defaults to None.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
            drop_last (bool): Drop the last batch if it is not complete. Defaults to True.
            num_workers (int): Number of parallel workers. Defaults to 0 for single threaded process.
            pin_memory (bool): If ``True``, the data loader will copy Tensors into device/CUDA pinned memory before 
                returning them. Defaults to False.
            data_with_sample_dim (bool): Use a specific collate function to concatenate samples along a existing sample
                dimension instead of stacking the samples. Defaults to False.
            sample_num_modalities (int, optional): Load only a subset of modalities per batch. Defaults to None.
            sample_replace (bool): If sample_num_modalities is set, sample modalities with replacement.
                Defaults to False.
            channel_position (int): Position of the channel dimension in the image modalities. Defaults to -3.
            concat_bands (bool): Concatenate all image modalities along the band dimension into a single "image", so
                that it can be processed by single-modal models. Concatenate in the order of provided modalities.
                Works with image modalities only. Does not work with allow_missing_modalities. Defaults to False.
        """
        if task == "segmentation":
            dataset_class = GenericMultimodalSegmentationDataset
        elif task == "regression":
            dataset_class = GenericMultimodalPixelwiseRegressionDataset
        elif task in ["classification", "multilabel_classification", "scalar_regression", "scalar"]:
            dataset_class = GenericMultimodalScalarDataset
            task = "scalar"
        elif task is None:
            dataset_class = GenericMultimodalDataset
        else:
            raise ValueError(f"Unknown task {task}, only segmentation and regression are supported.")

        super().__init__(dataset_class, batch_size, num_workers, **kwargs)
        self.num_classes = num_classes
        self.class_names = class_names
        self.modalities = modalities
        self.image_modalities = image_modalities or modalities
        self.non_image_modalities = list(set(self.modalities) - set(self.image_modalities))
        if task == "scalar":
            self.non_image_modalities += ["label"]
        if isinstance(image_grep, dict):
            self.image_grep = {m: image_grep[m] if m in image_grep else "*" for m in modalities}
        else:
            self.image_grep = {m: image_grep or "*" for m in modalities}
        self.label_grep = label_grep or "*"
        self.train_root = train_data_root
        self.val_root = val_data_root
        self.test_root = test_data_root
        self.train_label_data_root = train_label_data_root
        self.val_label_data_root = val_label_data_root
        self.test_label_data_root = test_label_data_root
        self.predict_root = predict_data_root
        self.train_split = train_split
        self.val_split = val_split
        self.test_split = test_split
        self.allow_substring_file_names = allow_substring_file_names
        self.constant_scale = constant_scale
        self.no_data_replace = no_data_replace
        self.no_label_replace = no_label_replace
        self.drop_last = drop_last
        self.pin_memory = pin_memory
        self.sample_num_modalities = sample_num_modalities
        self.sample_replace = sample_replace        

        self.dataset_bands = dataset_bands
        self.output_bands = output_bands
        self.predict_dataset_bands = predict_dataset_bands
        self.predict_output_bands = predict_output_bands

        self.rgb_modality = rgb_modality or modalities[0]
        self.rgb_indices = rgb_indices
        self.expand_temporal_dimension = expand_temporal_dimension
        self.reduce_zero_label = reduce_zero_label
        self.channel_position = channel_position
        self.concat_bands = concat_bands

        if isinstance(train_transform, dict):
            self.train_transform = {m: wrap_in_compose_is_list(train_transform[m]) if m in train_transform else None
                                    for m in modalities}
        elif shared_transforms:
            self.train_transform = wrap_in_compose_is_list(train_transform,
                                                           image_modalities=self.image_modalities,
                                                           non_image_modalities=self.non_image_modalities)
        else:
            self.train_transform = {m: wrap_in_compose_is_list(train_transform)
                                    for m in modalities}

        if isinstance(val_transform, dict):
            self.val_transform = {m: wrap_in_compose_is_list(val_transform[m]) if m in val_transform else None
                                    for m in modalities}
        elif shared_transforms:
            self.val_transform = wrap_in_compose_is_list(val_transform,
                                                         image_modalities=self.image_modalities,
                                                         non_image_modalities=self.non_image_modalities)
        else:
            self.val_transform = {m: wrap_in_compose_is_list(val_transform)
                                    for m in modalities}

        if isinstance(test_transform, dict):
            self.test_transform = {m: wrap_in_compose_is_list(test_transform[m]) if m in test_transform else None
                                    for m in modalities}
        elif shared_transforms:
            self.test_transform = wrap_in_compose_is_list(test_transform,
                                                          image_modalities=self.image_modalities,
                                                          non_image_modalities=self.non_image_modalities,                                                          
                                                          )
        else:
            self.test_transform = {m: wrap_in_compose_is_list(test_transform)
                                    for m in modalities}

        if self.concat_bands:
            # Concatenate mean and std values
            means = load_from_file_or_attribute(np.concatenate([means[m] for m in self.image_modalities]).tolist())
            stds = load_from_file_or_attribute(np.concatenate([stds[m] for m in self.image_modalities]).tolist())

            self.aug = Normalize(means, stds)
        else:
            # Apply standardization per modality
            means = {m: load_from_file_or_attribute(means[m]) for m in means.keys()}
            stds = {m: load_from_file_or_attribute(stds[m]) for m in stds.keys()}

            self.aug = MultimodalNormalize(means, stds)

        self.data_with_sample_dim = data_with_sample_dim

        self.collate_fn = collate_chunk_dicts if data_with_sample_dim else collate_samples


    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(
                data_root=self.train_root,
                num_classes=self.num_classes,
                image_grep=self.image_grep,
                label_grep=self.label_grep,
                label_data_root=self.train_label_data_root,
                split=self.train_split,
                allow_substring_file_names=self.allow_substring_file_names,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                image_modalities=self.image_modalities,
                rgb_modality=self.rgb_modality,
                rgb_indices=self.rgb_indices,
                transform=self.train_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
                channel_position=self.channel_position,
                concat_bands=self.concat_bands ,
            )
            logger.info(f"Train dataset: {len(self.train_dataset)}")
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                data_root=self.val_root,
                num_classes=self.num_classes,
                image_grep=self.image_grep,
                label_grep=self.label_grep,
                label_data_root=self.val_label_data_root,
                split=self.val_split,
                allow_substring_file_names=self.allow_substring_file_names,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                image_modalities=self.image_modalities,
                rgb_modality=self.rgb_modality,
                rgb_indices=self.rgb_indices,
                transform=self.val_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
                channel_position=self.channel_position,
                concat_bands=self.concat_bands,
            )
            logger.info(f"Val dataset: {len(self.val_dataset)}")
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                data_root=self.test_root,
                num_classes=self.num_classes,
                image_grep=self.image_grep,
                label_grep=self.label_grep,
                label_data_root=self.test_label_data_root,
                split=self.test_split,
                allow_substring_file_names=self.allow_substring_file_names,
                dataset_bands=self.dataset_bands,
                output_bands=self.output_bands,
                constant_scale=self.constant_scale,
                image_modalities=self.image_modalities,
                rgb_modality=self.rgb_modality,
                rgb_indices=self.rgb_indices,
                transform=self.test_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
                channel_position=self.channel_position,
                concat_bands=self.concat_bands,
            )
            logger.info(f"Test dataset: {len(self.test_dataset)}")
        if stage in ["predict"] and self.predict_root:
            self.predict_dataset = self.dataset_class(
                data_root=self.predict_root,
                num_classes=self.num_classes,
                allow_substring_file_names=self.allow_substring_file_names,
                dataset_bands=self.predict_dataset_bands,
                output_bands=self.predict_output_bands,
                constant_scale=self.constant_scale,
                image_modalities=self.image_modalities,
                rgb_modality=self.rgb_modality,
                rgb_indices=self.rgb_indices,
                transform=self.test_transform,
                no_data_replace=self.no_data_replace,
                no_label_replace=self.no_label_replace,
                expand_temporal_dimension=self.expand_temporal_dimension,
                reduce_zero_label=self.reduce_zero_label,
                channel_position=self.channel_position,
                concat_bands=self.concat_bands,
            )
            logger.info(f"Predict dataset: {len(self.predict_dataset)}")

    def _dataloader_factory(self, split: str) -> DataLoader[dict[str, Tensor]]:
        """Implement one or more PyTorch DataLoaders.

        Args:
            split: Either "train", "val", "test", or "predict".

        Returns:
            A collection of data loaders specifying samples.

        Raises:
            MisconfigurationException: If :meth:`setup` does not define a
                dataset or sampler, or if the dataset or sampler has length 0.
        """
        dataset = self._valid_attribute(f"{split}_dataset", "dataset")
        batch_size = self._valid_attribute(f"{split}_batch_size", "batch_size")
        if self.sample_num_modalities:
            # Custom batch sampler for sampling modalities per batch
            batch_sampler = MultiModalBatchSampler(
                self.modalities, self.sample_num_modalities, self.sample_replace,
                RandomSampler(dataset) if split == "train" else SequentialSampler(dataset),
                batch_size=batch_size,
                drop_last=split == "train" and self.drop_last
            )
        else:
            batch_sampler = BatchSampler(
                RandomSampler(dataset) if split == "train" else SequentialSampler(dataset),
                batch_size=batch_size,
                drop_last=split == "train" and self.drop_last
            )

        return DataLoader(
            dataset=dataset,
            batch_sampler=batch_sampler,
            num_workers=self.num_workers,
            collate_fn=self.collate_fn,
            pin_memory=self.pin_memory,
        )



# Copyright contributors to the Terratorch project
"""
This module should be removed when PrithviModelFactory is removed. For now, this tests backwards compatibility.
"""
import pytest
import torch

from terratorch.models import PrithviModelFactory
from terratorch.models.backbones.prithvi_vit import PRETRAINED_BANDS
from terratorch.models.model import AuxiliaryHead
import gc 

NUM_CHANNELS = 6
NUM_CLASSES = 2
EXPECTED_SEGMENTATION_OUTPUT_SHAPE = (1, NUM_CLASSES, 224, 224)
EXPECTED_REGRESSION_OUTPUT_SHAPE = (1, 224, 224)
EXPECTED_CLASSIFICATION_OUTPUT_SHAPE = (1, NUM_CLASSES)

PIXELWISE_TASK_EXPECTED_OUTPUT = [
    ("regression", EXPECTED_REGRESSION_OUTPUT_SHAPE),
    ("segmentation", EXPECTED_SEGMENTATION_OUTPUT_SHAPE),
]

@pytest.fixture(scope="session")
def model_factory() -> PrithviModelFactory:
    return PrithviModelFactory()


@pytest.fixture(scope="session")
def model_input() -> torch.Tensor:
    return torch.ones((1, NUM_CHANNELS, 224, 224))


@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100", "prithvi_eo_v2_300"])
def test_create_classification_model(backbone, model_factory: PrithviModelFactory, model_input):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        in_channels=NUM_CHANNELS,
        bands=PRETRAINED_BANDS,
        pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE


@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100", "prithvi_eo_v2_300"])
def test_create_classification_model_no_in_channels(backbone, model_factory: PrithviModelFactory, model_input):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        bands=PRETRAINED_BANDS,
        pretrained=False,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

    gc.collect()

@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_pixelwise_model(backbone, task, expected, decoder, model_factory: PrithviModelFactory, model_input):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "in_channels": NUM_CHANNELS,
        "bands": PRETRAINED_BANDS,
        "pretrained": False,
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
    if decoder == "UperNetDecoder":
        model_args["backbone_out_indices"] = [1, 2, 3, 4]
        model_args["decoder_scale_modules"] = True

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()

@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_pixelwise_model_no_in_channels(
    backbone, task, expected, decoder, model_factory: PrithviModelFactory, model_input
):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "bands": PRETRAINED_BANDS,
        "pretrained": False,
    }

    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES
    if decoder == "UperNetDecoder":
        model_args["backbone_out_indices"] = [1, 2, 3, 4]
        model_args["decoder_scale_modules"] = True

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()

@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_pixelwise_model_with_aux_heads(
    backbone, task, expected, decoder, model_factory: PrithviModelFactory, model_input
):
    aux_heads_name = ["first_aux", "second_aux"]
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "in_channels": NUM_CHANNELS,
        "bands": PRETRAINED_BANDS,
        "pretrained": False,
        "aux_decoders": [AuxiliaryHead(name, "FCNDecoder", None) for name in aux_heads_name],
    }
    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES

    if decoder == "UperNetDecoder":
        model_args["backbone_out_indices"] = [1, 2, 3, 4]
        model_args["decoder_scale_modules"] = True

    model = model_factory.build_model(**model_args)
    model.eval()

    with torch.no_grad():
        model_output = model(model_input)
        assert model_output.output.shape == expected

        assert len(model_output.auxiliary_heads.keys() & aux_heads_name) == len(aux_heads_name)
        for _, output in model_output.auxiliary_heads.items():
            assert output.shape == expected

    gc.collect()

@pytest.mark.parametrize("backbone", ["prithvi_eo_v1_100"])
@pytest.mark.parametrize("task,expected", PIXELWISE_TASK_EXPECTED_OUTPUT)
@pytest.mark.parametrize("decoder", ["FCNDecoder", "UperNetDecoder", "IdentityDecoder"])
def test_create_pixelwise_model_with_extra_bands(backbone, task, expected, decoder, model_factory: PrithviModelFactory):
    model_args = {
        "task": task,
        "backbone": backbone,
        "decoder": decoder,
        "in_channels": NUM_CHANNELS + 1,
        "bands": [*PRETRAINED_BANDS, 7],
        "pretrained": False,
    }
    if task == "segmentation":
        model_args["num_classes"] = NUM_CLASSES

    if decoder == "UperNetDecoder":
        model_args["backbone_out_indices"] = [1, 2, 3, 4]
        model_args["decoder_scale_modules"] = True
    model = model_factory.build_model(**model_args)
    model.eval()
    model_input = torch.ones((1, NUM_CHANNELS + 1, 224, 224))
    with torch.no_grad():
        assert model(model_input).output.shape == expected

    gc.collect()



# reference torchgeo https://torchgeo.readthedocs.io/en/latest/_modules/torchgeo/models/dofa.html#DOFA
import torch
import torchgeo.models.dofa as dofa
import logging
from collections.abc import Callable
from functools import partial
import huggingface_hub
import torch.nn as nn
from typing import List
import huggingface_hub
from torchvision.models._api import Weights, WeightsEnum
from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY
import pdb

waves_list= {
  "COASTAL_AEROSOL": 0.44,
  "BLUE": 0.49,
  "GREEN": 0.56,
  "RED": 0.665,
  "RED_EDGE_1": 0.705,
  "RED_EDGE_2": 0.74, 
  "RED_EDGE_3": 0.783,
  "NIR_BROAD": 0.832,
  "NIR_NARROW": 0.864,
  "WATER_VAPOR": 0.945,
  "CIRRUS": 1.373,
  "SWIR_1": 1.61,
  "SWIR_2": 2.20,
  "THEMRAL_INFRARED_1": 10.90,
  "THEMRAL_INFRARED_12": 12.00, 
  "VV": 3.75,
  "VH": 3.75,
  "ASC_VV": 3.75,
  "ASC_VH": 3.75,
  "DSC_VV": 3.75,
  "DSC_VH": 3.75,
  "VV-VH": 3.75
}


class DOFAEncoderWrapper(nn.Module):

    """
    A wrapper for DOFA models from torchgeo to return only the forward pass of the encoder 
    Attributes:
        dofa_model (DOFA): The instantiated dofa model
    Methods:
        forward(x: List[torch.Tensor], wavelengths: list[float]) -> torch.Tensor:
            Forward pass for embeddings with specified indices.
    """

    def __init__(self, dofa_model, wavelengths, weights=None, out_indices=None) -> None:
        """
        Args:
            dofa_model (DOFA): The decoder module to be wrapped.
            weights ()
        """
        super().__init__()
        self.dofa_model = dofa_model
        self.weights = weights
        self.wavelengths = wavelengths

        self.out_indices = out_indices if out_indices else [-1]
        self.out_channels = [self.dofa_model.patch_embed.embed_dim] * len(self.out_indices)
        
    def forward(self, x: List[torch.Tensor], **kwargs) -> torch.Tensor:
        
        wavelist = torch.tensor(self.wavelengths, device=x.device).float()

        x, _ = self.dofa_model.patch_embed(x, wavelist)

        x = x + self.dofa_model.pos_embed[:, 1:, :]
        # append cls token
        cls_token = self.dofa_model.cls_token + self.dofa_model.pos_embed[:, :1, :]
        cls_tokens = cls_token.expand(x.shape[0], -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)

        outs = []
        # apply Transformer blocks
        for i, block in enumerate(self.dofa_model.blocks):
            x = block(x)
            if i in self.out_indices:
                outs.append(x)
            elif (i == (len(self.dofa_model.blocks)-1)) & (-1 in self.out_indices):
                outs.append(x)
                
        return tuple(outs)

def get_wavelenghts(model_bands):

    wavelengths = [waves_list[x.split('.')[-1]] for x in model_bands]

    return wavelengths
    

@TERRATORCH_BACKBONE_REGISTRY.register
def dofa_small_patch16_224(model_bands, input_size = 224, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = None, out_indices: list | None = None, **kwargs):
    model = dofa.dofa_small_patch16_224(**kwargs)
    input_size = kwargs["img_size"] if "img_size" in kwargs else 224
    if pretrained:
        model = load_dofa_weights(model, ckpt_data, weights, input_size)
    wavelengths = get_wavelenghts(model_bands)
    
    return DOFAEncoderWrapper(model, wavelengths, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def dofa_base_patch16_224(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = dofa.DOFABase16_Weights.DOFA_MAE, out_indices: list | None = None, **kwargs):
    model = dofa.dofa_base_patch16_224(**kwargs)
    input_size = kwargs["img_size"] if "img_size" in kwargs else 224
    if pretrained:
        model = load_dofa_weights(model, ckpt_data, weights, input_size)
    wavelengths = get_wavelenghts(model_bands)
    
    return DOFAEncoderWrapper(model, wavelengths, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def dofa_large_patch16_224(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = dofa.DOFALarge16_Weights.DOFA_MAE, out_indices: list | None = None, **kwargs):
    model = dofa.dofa_large_patch16_224(**kwargs)
    input_size = kwargs["img_size"] if "img_size" in kwargs else 224
    if pretrained:
        model = load_dofa_weights(model, ckpt_data, weights, input_size)
    wavelengths = get_wavelenghts(model_bands)
    
    return DOFAEncoderWrapper(model, wavelengths, weights, out_indices)

@TERRATORCH_BACKBONE_REGISTRY.register
def dofa_huge_patch16_224(model_bands, pretrained = False, ckpt_data: str | None = None,  weights: Weights | None = None, out_indices: list | None = None, **kwargs):
    model = dofa.dofa_huge_patch16_224(**kwargs)
    input_size = kwargs["img_size"] if "img_size" in kwargs else 224
    if pretrained:
        model = load_dofa_weights(model, ckpt_data, weights, input_size)
    wavelengths = get_wavelenghts(model_bands)
    
    return DOFAEncoderWrapper(model, wavelengths, weights, out_indices)

def load_dofa_weights(model: nn.Module, ckpt_data: str | None = None,  weights: Weights | None = None, input_size = 224) -> nn.Module:
    state_dict = model.state_dict()
    print("Loading weights")
    if ckpt_data is not None:
        if ckpt_data.find("https://hf.co/") > -1:
            repo_id = ckpt_data.split("/resolve/")[0].replace("https://hf.co/", '')
            filename = ckpt_data.split("/")[-1]
            ckpt_data = huggingface_hub.hf_hub_download(repo_id=repo_id, filename=filename)
        checkpoint_model = torch.load(ckpt_data, map_location="cpu")

        for k in ["head.weight", "head.bias"]:
                if (
                    k in checkpoint_model
                    and checkpoint_model[k].shape != state_dict[k].shape
                ):
                    logging.info(f"Removing key {k} from pretrained checkpoint")
                    del checkpoint_model[k]
        if input_size != 224:
            if (
                "pos_embed" in checkpoint_model
                and checkpoint_model["pos_embed"].shape != state_dict["pos_embed"].shape
            ):
                logging.info("Removing key pos_embed from pretrained checkpoint")
                del checkpoint_model["pos_embed"]

    
        msg = model.load_state_dict(checkpoint_model, strict=False)
    
        logging.info(msg)
    else:
        if weights is not None:
            
            checkpoint_model = weights.get_state_dict(progress=True)
            allowed_missing_keys =  {'fc_norm.weight', 'fc_norm.bias', 'head.weight', 'head.bias'}
            if input_size != 224:
                if (
                    "pos_embed" in checkpoint_model
                    and checkpoint_model["pos_embed"].shape != state_dict["pos_embed"].shape
                ):
                    logging.info("Removing key pos_embed from pretrained checkpoint")
                    del checkpoint_model["pos_embed"]
                    allowed_missing_keys.add('pos_embed')
            missing_keys, unexpected_keys = model.load_state_dict(checkpoint_model, strict=False)
            logging.info(f"Weights loaded.")
            # Both fc_norm and head are generated dynamically
            assert set(missing_keys) <= allowed_missing_keys
            assert not unexpected_keys
        else:
            print("No weights to load.")
            
    return model




# Copyright contributors to the Terratorch project

from terratorch.datamodules.fire_scars import FireScarsNonGeoDataModule
from terratorch.datamodules.generic_pixel_wise_data_module import (
    GenericNonGeoPixelwiseRegressionDataModule,
    GenericNonGeoSegmentationDataModule,
)
from terratorch.datamodules.generic_scalar_label_data_module import (
    GenericNonGeoClassificationDataModule,
)
from terratorch.datamodules.m_bigearthnet import MBigEarthNonGeoDataModule
from terratorch.datamodules.m_brick_kiln import MBrickKilnNonGeoDataModule

# geobench segmentation datamodules
from terratorch.datamodules.m_cashew_plantation import MBeninSmallHolderCashewsNonGeoDataModule
from terratorch.datamodules.m_chesapeake_landcover import MChesapeakeLandcoverNonGeoDataModule

# geobench classification datamodules
from terratorch.datamodules.m_eurosat import MEuroSATNonGeoDataModule
from terratorch.datamodules.m_forestnet import MForestNetNonGeoDataModule
from terratorch.datamodules.m_neontree import MNeonTreeNonGeoDataModule
from terratorch.datamodules.m_nz_cattle import MNzCattleNonGeoDataModule
from terratorch.datamodules.m_pv4ger import MPv4gerNonGeoDataModule
from terratorch.datamodules.m_pv4ger_seg import MPv4gerSegNonGeoDataModule
from terratorch.datamodules.m_SA_crop_type import MSACropTypeNonGeoDataModule
from terratorch.datamodules.m_so2sat import MSo2SatNonGeoDataModule
from terratorch.datamodules.multi_temporal_crop_classification import MultiTemporalCropClassificationDataModule
from terratorch.datamodules.open_sentinel_map import OpenSentinelMapDataModule
from terratorch.datamodules.pastis import PASTISDataModule
from terratorch.datamodules.era5 import ERA5DataModule

try:
    wxc_present = True
    from terratorch.datamodules.merra2_downscale import Merra2DownscaleNonGeoDataModule 
except ImportError as e:
    import logging
    logging.getLogger("terratorch").debug("wxc_downscaling not installed")
    wxc_present = False

# GenericNonGeoRegressionDataModule,
from terratorch.datamodules.sen1floods11 import Sen1Floods11NonGeoDataModule
from terratorch.datamodules.sen4agrinet import Sen4AgriNetDataModule
from terratorch.datamodules.torchgeo_data_module import TorchGeoDataModule, TorchNonGeoDataModule
from terratorch.datamodules.generic_multimodal_data_module import GenericMultiModalDataModule


# miscellaneous datamodules
from terratorch.datamodules.openearthmap import OpenEarthMapNonGeoDataModule

from terratorch.datamodules.burn_intensity import BurnIntensityNonGeoDataModule
from terratorch.datamodules.carbonflux import CarbonFluxNonGeoDataModule
from terratorch.datamodules.landslide4sense import Landslide4SenseNonGeoDataModule
from terratorch.datamodules.biomassters import BioMasstersNonGeoDataModule
from terratorch.datamodules.forestnet import ForestNetNonGeoDataModule

# miscellaneous datamodules
from terratorch.datamodules.openearthmap import OpenEarthMapNonGeoDataModule

# Generic classification datamodule
from terratorch.datamodules.sen4map import Sen4MapLucasDataModule

__all__ = (
    "GenericNonGeoSegmentationDataModule",
    "GenericNonGeoPixelwiseRegressionDataModule",
    "GenericNonGeoSegmentationDataModule",
    "GenericNonGeoClassificationDataModule",
    # "GenericNonGeoRegressionDataModule",
    "BurnIntensityNonGeoDataModule",
    "CarbonFluxNonGeoDataModule",
    "Landslide4SenseNonGeoDataModule",
    "ForestNetNonGeoDataModule",
    "BioMasstersNonGeoDataModule"
    "Sen1Floods11NonGeoDataModule",
    "Sen4MapLucasDataModule",
    "FireScarsNonGeoDataModule",
    "MultiTemporalCropClassificationDataModule",
    "TorchGeoDataModule",
    "TorchNonGeoDataModule",
    "MEuroSATNonGeoDataModule",
    "MBigEarthNonGeoDataModule",
    "MBrickKilnNonGeoDataModule",
    "MForestNetNonGeoDataModule",
    "MSo2SatNonGeoDataModule",
    "MPv4gerNonGeoDataModule",
    "MBeninSmallHolderCashewsNonGeoDataModule",
    "MNzCattleNonGeoDataModule",
    "MChesapeakeLandcoverNonGeoDataModule",
    "MPv4gerSegNonGeoDataModule",
    "MSACropTypeNonGeoDataModule",
    "MNeonTreeNonGeoDataModule",
    "OpenEarthMapModule"
    "OpenSentinelMapDataModule",
    "PASTISDataModule",
    "Sen4AgriNetDataModule",
    "GenericMultiModalDataModule",
)

if wxc_present:
    __all__.__add__(("Merra2DownscaleNonGeoDataModule", ))


# Copyright contributors to the Terratorch project

from terratorch.models.heads.classification_head import ClassificationHead
from terratorch.models.heads.regression_head import RegressionHead
from terratorch.models.heads.segmentation_head import SegmentationHead

__all__ = ["ClassificationHead", "RegressionHead", "SegmentationHead"]



import typing
from collections import OrderedDict
from collections.abc import Callable, Mapping, Set
from contextlib import suppress
from reprlib import recursive_repr as _recursive_repr

class BuildableRegistry(typing.Protocol):
    def __iter__(self): ...
    def __len__(self) -> int: ...
    def __contains__(self, name: str) -> bool: ...
    def build(self, name: str, *args, **kwargs):
        """Should raise KeyError if model not found.
        """


class DecoderRegistry(BuildableRegistry, typing.Protocol):
    includes_head: bool

T = typing.TypeVar("T", bound=BuildableRegistry)
class MultiSourceRegistry(Mapping[str, T], typing.Generic[T]):
    """Registry that searches in multiple sources

        Correct functioning of this class depends on registries raising a KeyError when the model is not found.
    """
    def __init__(self, **sources) -> None:
        self._sources: OrderedDict[str, T] = OrderedDict(sources)

    def _parse_prefix(self, name) -> tuple[str, str] | None:
        split = name.split("_")
        if len(split) > 1 and split[0] in self._sources:
            prefix = split[0]
            name_without_prefix = "_".join(split[1:])
            return prefix, name_without_prefix
        return None

    def find_registry(self, name: str) -> T:
        parsed_prefix = self._parse_prefix(name)
        if parsed_prefix:
            prefix, name_without_prefix = parsed_prefix
            registry = self._sources[prefix]
            return registry

        # if no prefix is given, go through all sources in order
        for registry in self._sources.values():
            if name in registry:
                return registry
        msg = f"Model {name} not found in any registry"
        raise KeyError(msg)

    def build(self, name: str, *constructor_args, **constructor_kwargs):
        parsed_prefix = self._parse_prefix(name)
        if parsed_prefix:
            prefix, name_without_prefix = parsed_prefix
            registry = self._sources[prefix]
            return registry.build(name_without_prefix, *constructor_args, **constructor_kwargs)

        # if no prefix, try to build in order
        for source in self._sources.values():
            with suppress(KeyError):
                return source.build(name, *constructor_args, **constructor_kwargs)

        msg = f"Could not instantiate model {name} not from any source."
        raise KeyError(msg)

    def register_source(self, prefix: str, registry: T) -> None:
        """Register a source in the registry"""
        if prefix in self._sources:
            msg = f"Source for prefix {prefix} already exists."
            raise KeyError(msg)
        self._sources[prefix] = registry

    def __iter__(self):
        for prefix in self._sources:
            for element in self._sources[prefix]:
                yield prefix + "_" + element

    def __len__(self):
        return sum(len(source) for source in self._sources.values())

    def __getitem__(self, name):
        return self._sources[name]

    def __contains__(self, name):
        parsed_prefix = self._parse_prefix(name)
        if parsed_prefix:
            prefix, name_without_prefix = parsed_prefix
            return name_without_prefix in self._sources[prefix]
        return any(name in source for source in self._sources.values())

    @_recursive_repr()
    def __repr__(self):
        args = [f"{name}={source!r}" for name, source in self._sources.items()]
        return f'{self.__class__.__name__}({", ".join(args)})'

    def __str__(self):
        sources_str = str(" | ".join([f"{prefix}: {source!s}" for prefix, source in self._sources.items()]))
        return f"Multi source registry with {len(self)} items: {sources_str}"

    def keys(self):
        return self._sources.keys()


class Registry(Set):
    """Registry holding model constructors and multiple additional sources.

    This registry behaves as a set of strings, which are model names,
    to model classes or functions which instantiate model classes.

    In addition, it can instantiate models with the build method.

    Add constructors to the registry by annotating them with @registry.register.

    >>> registry = Registry()
    >>> @registry.register
    ... def model(*args, **kwargs):
    ...     return object()
    >>> "model" in registry
    True
    >>> model_instance = registry.build("model")
    """

    def __init__(self, **elements) -> None:
        self._registry: dict[str, Callable] = dict(elements)

    def register(self, constructor: Callable | type) -> Callable:
        """Register a component in the registry. Used as a decorator.

        Args:
            constructor (Callable | type): Function or class to be decorated with @register.
        """
        if not callable(constructor):
            msg = f"Invalid argument. Decorate a function or class with @{self.__class__.__name__}.register"
            raise TypeError(msg)
        self._registry[constructor.__name__] = constructor
        return constructor

    def build(self, name: str, *constructor_args, **constructor_kwargs):
        """Build and return the component.
        Use prefixes ending with _ to forward to a specific source
        """
        return self._registry[name](*constructor_args, **constructor_kwargs)

    def __iter__(self):
        return iter(self._registry)

    # def __getitem__(self, key):
    #     return self._registry[key]

    def __len__(self):
        return len(self._registry)

    def __contains__(self, key):
        return key in self._registry

    def __repr__(self):
        return f"{self.__class__.__name__}({self._registry!r})"

    def __str__(self):
        return f"Registry with {len(self)} registered items"

## Declare library level registries below

### Backbone Registry
TERRATORCH_BACKBONE_REGISTRY = Registry()
BACKBONE_REGISTRY: MultiSourceRegistry[BuildableRegistry] = MultiSourceRegistry()
BACKBONE_REGISTRY.register_source("terratorch", TERRATORCH_BACKBONE_REGISTRY)

### Neck Registry
NECK_REGISTRY: MultiSourceRegistry[BuildableRegistry] = MultiSourceRegistry()
TERRATORCH_NECK_REGISTRY = Registry()
NECK_REGISTRY.register_source("terratorch", TERRATORCH_NECK_REGISTRY)

### Decoder Registry
TERRATORCH_DECODER_REGISTRY = typing.cast(DecoderRegistry, Registry())
TERRATORCH_DECODER_REGISTRY.includes_head = False
DECODER_REGISTRY: MultiSourceRegistry[DecoderRegistry] = MultiSourceRegistry()
DECODER_REGISTRY.register_source("terratorch", TERRATORCH_DECODER_REGISTRY)

### Model Factory Registry
MODEL_FACTORY_REGISTRY = Registry()



# Copyright contributors to the Terratorch project

"""This module handles registering prithvi_swin models into timm.
"""

import logging
import warnings
from collections import OrderedDict

import torch
from timm.models import SwinTransformer
from timm.models._builder import build_model_with_cfg
from timm.models._registry import generate_default_cfgs, register_model

from terratorch.datasets.utils import HLSBands
from terratorch.models.backbones.select_patch_embed_weights import select_patch_embed_weights
from terratorch.models.backbones.swin_encoder_decoder import MMSegSwinTransformer
from terratorch.datasets.utils import generate_bands_intervals

logger = logging.getLogger(__name__)

PRETRAINED_BANDS = [
    HLSBands.BLUE,
    HLSBands.GREEN,
    HLSBands.RED,
    HLSBands.NIR_NARROW,
    HLSBands.SWIR_1,
    HLSBands.SWIR_2,
]


default_cfgs = generate_default_cfgs(
    {
        "prithvi_swin_B": {},
        "prithvi_swin_L": {},
    }
)

def convert_weights_swin2mmseg(ckpt):
    # from https://github.com/open-mmlab/mmsegmentation/blob/main/tools/model_converters/swin2mmseg.py
    new_ckpt = OrderedDict()

    def correct_unfold_reduction_order(x):
        out_channel, in_channel = x.shape
        x = x.reshape(out_channel, 4, in_channel // 4)
        x = x[:, [0, 2, 1, 3], :].transpose(1, 2).reshape(out_channel, in_channel)
        return x

    def correct_unfold_norm_order(x):
        in_channel = x.shape[0]
        x = x.reshape(4, in_channel // 4)
        x = x[[0, 2, 1, 3], :].transpose(0, 1).reshape(in_channel)
        return x

    for k, v in ckpt.items():
        if k.startswith("head"):
            continue
        elif k.startswith("layers"):
            new_v = v
            if "attn." in k:
                new_k = k.replace("attn.", "attn.w_msa.")
            elif "mlp." in k:
                if "mlp.fc1." in k:
                    new_k = k.replace("mlp.fc1.", "ffn.layers.0.0.")
                elif "mlp.fc2." in k:
                    new_k = k.replace("mlp.fc2.", "ffn.layers.1.")
                else:
                    new_k = k.replace("mlp.", "ffn.")
            elif "downsample" in k:
                new_k = k
                if "reduction." in k:
                    new_v = correct_unfold_reduction_order(v)
                elif "norm." in k:
                    new_v = correct_unfold_norm_order(v)
            else:
                new_k = k
            new_k = new_k.replace("layers", "stages", 1)
        elif k.startswith("patch_embed"):
            new_v = v
            if "proj" in k:
                new_k = k.replace("proj", "projection")
            else:
                new_k = k
        else:
            new_v = v
            new_k = k

        new_ckpt[new_k] = new_v

    return new_ckpt


def weights_are_swin_implementation(state_dict: dict[str, torch.Tensor]):
    # if keys start with 'encoder', treat it as the swin implementation
    for k in state_dict.keys():
        if k.startswith("encoder."):
            return True
    return False


def checkpoint_filter_fn(state_dict: dict[str, torch.Tensor], model: torch.nn.Module, pretrained_bands, model_bands):
    """convert patch embedding weight from manual patchify + linear proj to conv"""
    if "head.fc.weight" in state_dict:
        return state_dict

    if "state_dict" in state_dict:
        _state_dict = state_dict["state_dict"]
    elif "model" in state_dict:
        _state_dict = state_dict["model"]
    else:
        _state_dict = state_dict

    # strip prefix of state_dict
    if next(iter(_state_dict.keys())).startswith("module."):
        _state_dict = {k[7:]: v for k, v in _state_dict.items()}

    if weights_are_swin_implementation(_state_dict):
        # keep only encoder weights
        state_dict = OrderedDict()
        for k, v in _state_dict.items():
            if k.startswith("encoder."):
                state_dict[k[8:]] = v
            elif not k.startswith("decoder"):
                state_dict[k] = v
        state_dict = convert_weights_swin2mmseg(state_dict)
    else:
        # keep only encoder weights
        state_dict = OrderedDict()

        for k, v in _state_dict.items():
            if k.startswith("backbone."):
                state_dict[k[9:]] = v
            else:
                state_dict[k] = v

    relative_position_bias_table_keys = [k for k in state_dict.keys() if "relative_position_bias_table" in k]
    for table_key in relative_position_bias_table_keys:
        table_pretrained = state_dict[table_key]
        table_current = model.state_dict()[table_key]
        L1, nH1 = table_pretrained.size()
        L2, nH2 = table_current.size()
        if nH1 != nH2:
            warnings.warn(f"Error in loading {table_key}, pass", stacklevel=1)
        elif L1 != L2:
            S1 = int(L1**0.5)
            S2 = int(L2**0.5)
            table_pretrained_resized = torch.nn.functional.interpolate(
                table_pretrained.permute(1, 0).reshape(1, nH1, S1, S1),
                size=(S2, S2),
                mode="bicubic",
            )
            state_dict[table_key] = table_pretrained_resized.view(nH2, L2).permute(1, 0).contiguous()

    if hasattr(model.head.fc, "weight"):
        state_dict["head.fc.weight"] = model.head.fc.weight.detach().clone()
        state_dict["head.fc.bias"] = model.head.fc.bias.detach().clone()

    state_dict = select_patch_embed_weights(state_dict, model, pretrained_bands, model_bands)
    return state_dict


def _create_swin_mmseg_transformer(
    variant: str,
    pretrained_bands: list[HLSBands] | None = None,
    model_bands: list[HLSBands | int] | None = None,
    pretrained: bool = False,  # noqa: FBT002, FBT001
    **kwargs,
):
    if pretrained_bands is None:
        pretrained_bands = PRETRAINED_BANDS

    if model_bands is None:
        model_bands: list[HLSBands | int] = pretrained_bands
        logger.info(
            f"Model bands not passed. Assuming bands are ordered in the same way as {PRETRAINED_BANDS}.\
            Pretrained patch_embed layer may be misaligned with current bands"
        )
    else:
        model_bands = [HLSBands.try_convert_to_hls_bands_enum(b) for b in model_bands]
    default_out_indices = tuple(i for i, _ in enumerate(kwargs.get("depths", (1, 1, 3, 1))))
    out_indices = kwargs.pop("out_indices", default_out_indices)

    # the current swin model is not multitemporal
    if "num_frames" in kwargs:
        kwargs = {k: v for k, v in kwargs.items() if k != "num_frames"}

    model_bands = generate_bands_intervals(model_bands)
    kwargs["in_chans"] = len(model_bands)

    def checkpoint_filter_wrapper_fn(state_dict, model):
        return checkpoint_filter_fn(state_dict, model, pretrained_bands, model_bands)

    # When the pretrained configuration is not available in HF, we shift to 
    # pretrained=False
    try:
        model: MMSegSwinTransformer = build_model_with_cfg(
            MMSegSwinTransformer,
            variant,
            pretrained,
            pretrained_filter_fn=checkpoint_filter_wrapper_fn,
            pretrained_strict=False,
            feature_cfg={"flatten_sequential": True, "out_indices": out_indices},
            **kwargs,
        )
    except RuntimeError:
        print(f"No pretrained configuration was found for the model {variant}.")
        model: MMSegSwinTransformer = build_model_with_cfg(
            MMSegSwinTransformer,
            variant,
            False,
            pretrained_filter_fn=checkpoint_filter_wrapper_fn,
            pretrained_strict=False,
            feature_cfg={"flatten_sequential": True, "out_indices": out_indices},
            **kwargs,
        )

    model.pretrained_bands = pretrained_bands
    model.model_bands = model_bands

    def prepare_features_for_image_model(x):
        return [
            layer_output.permute(0, 3, 1, 2).contiguous()
            for layer_output in x
        ]

    # add permuting here
    model.prepare_features_for_image_model = prepare_features_for_image_model
    return model


@register_model
def prithvi_swin_B(
    pretrained: bool = False,  # noqa: FBT002, FBT001
    pretrained_bands: list[HLSBands] | None = None,
    bands: list[HLSBands | int] | None = None,
    **kwargs,
) -> SwinTransformer:
    """Prithvi Swin B"""
    if pretrained_bands is None:
        pretrained_bands = PRETRAINED_BANDS
    if bands is None:
        bands = pretrained_bands
        logger.info(
            f"Model bands not passed. Assuming bands are ordered in the same way as {PRETRAINED_BANDS}.\
            Pretrained patch_embed layer may be misaligned with current bands"
        )

    model_args = {
        "patch_size": 4,
        "window_size": 7,
        "embed_dim": 128,
        "depths": (2, 2, 18, 2),
        "in_chans": 6,
        "num_heads": (4, 8, 16, 32),
    }
    transformer = _create_swin_mmseg_transformer(
        "prithvi_swin_B", pretrained_bands, bands, pretrained=pretrained, **dict(model_args, **kwargs)
    )
    return transformer


@register_model
def prithvi_swin_L(
    pretrained: bool = False,  # noqa: FBT002, FBT001
    pretrained_bands: list[HLSBands] | None = None,
    bands: list[HLSBands | int] | None = None,
    **kwargs,
) -> SwinTransformer:
    """Prithvi Swin L"""
    if pretrained_bands is None:
        pretrained_bands = PRETRAINED_BANDS
    if bands is None:
        bands = pretrained_bands
        logger.info(
            f"Model bands not passed. Assuming bands are ordered in the same way as {PRETRAINED_BANDS}.\
            Pretrained patch_embed layer may be misaligned with current bands"
        )

    model_args = {
        "patch_size": 4,
        "window_size": 7,
        "embed_dim": 192,
        "depths": (2, 2, 18, 2),
        "in_chans": 6,
        "num_heads": (6, 12, 24, 48),
    }
    transformer = _create_swin_mmseg_transformer(
        "prithvi_swin_L", pretrained_bands, bands, pretrained=pretrained, **dict(model_args, **kwargs)
    )
    return transformer



import glob
import os
from collections.abc import Sequence
from pathlib import Path
from typing import Any

import albumentations as A
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rioxarray
import torch
from einops import rearrange
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from torch import Tensor
from torchgeo.datasets import NonGeoDataset
from xarray import DataArray

from terratorch.datasets.utils import clip_image, default_transform, filter_valid_files, validate_bands


class MultiTemporalCropClassification(NonGeoDataset):
    """NonGeo dataset implementation for multi-temporal crop classification."""

    all_band_names = (
        "BLUE",
        "GREEN",
        "RED",
        "NIR_NARROW",
        "SWIR_1",
        "SWIR_2",
    )

    class_names = (
        "Natural Vegetation",
        "Forest",
        "Corn",
        "Soybeans",
        "Wetlands",
        "Developed / Barren",
        "Open Water",
        "Winter Wheat",
        "Alfalfa",
        "Fallow / Idle Cropland",
        "Cotton",
        "Sorghum",
        "Other",
    )

    rgb_bands = ("RED", "GREEN", "BLUE")

    BAND_SETS = {"all": all_band_names, "rgb": rgb_bands}

    num_classes = 13
    time_steps = 3
    splits = {"train": "training", "val": "validation"}  # Only train and val splits available
    metadata_file_name = "chip_df_final.csv"
    col_name = "chip_id"
    date_columns = ["first_img_date", "middle_img_date", "last_img_date"]

    def __init__(
        self,
        data_root: str,
        split: str = "train",
        bands: Sequence[str] = BAND_SETS["all"],
        transform: A.Compose | None = None,
        no_data_replace: float | None = None,
        no_label_replace: int | None = None,
        expand_temporal_dimension: bool = True,
        reduce_zero_label: bool = True,
        use_metadata: bool = False,
    ) -> None:
        """Constructor

        Args:
            data_root (str): Path to the data root directory.
            split (str): one of 'train' or 'val'.
            bands (list[str]): Bands that should be output by the dataset. Defaults to all bands.
            transform (A.Compose | None): Albumentations transform to be applied.
                Should end with ToTensorV2(). If used through the corresponding data module,
                should not include normalization. Defaults to None, which applies ToTensorV2().
            no_data_replace (float | None): Replace nan values in input images with this value.
                If None, does no replacement. Defaults to None.
            no_label_replace (int | None): Replace nan values in label with this value.
                If none, does no replacement. Defaults to None.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Defaults to True.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to True.
            use_metadata (bool): whether to return metadata info (time and location).
        """
        super().__init__()
        if split not in self.splits:
            msg = f"Incorrect split '{split}', please choose one of {self.splits}."
            raise ValueError(msg)
        split_name = self.splits[split]
        self.split = split

        validate_bands(bands, self.all_band_names)
        self.bands = bands
        self.band_indices = np.asarray([self.all_band_names.index(b) for b in bands])
        self.data_root = Path(data_root)

        data_dir = self.data_root / f"{split_name}_chips"
        self.image_files = sorted(glob.glob(os.path.join(data_dir, "*_merged.tif")))
        self.segmentation_mask_files = sorted(glob.glob(os.path.join(data_dir, "*.mask.tif")))
        split_file = data_dir / f"{split_name}_data.txt"

        with open(split_file) as f:
            split = f.readlines()
        valid_files = {rf"{substring.strip()}" for substring in split}
        self.image_files = filter_valid_files(
            self.image_files,
            valid_files=valid_files,
            ignore_extensions=True,
            allow_substring=True,
        )
        self.segmentation_mask_files = filter_valid_files(
            self.segmentation_mask_files,
            valid_files=valid_files,
            ignore_extensions=True,
            allow_substring=True,
        )

        self.no_data_replace = no_data_replace
        self.no_label_replace = no_label_replace
        self.reduce_zero_label = reduce_zero_label
        self.expand_temporal_dimension = expand_temporal_dimension
        self.use_metadata = use_metadata
        self.metadata = None
        if self.use_metadata:
            metadata_file = self.data_root / self.metadata_file_name
            self.metadata = pd.read_csv(metadata_file)
            self._build_image_metadata_mapping()

        # If no transform is given, apply only to transform to torch tensor
        self.transform = transform if transform else default_transform

    def _build_image_metadata_mapping(self):
        """Build a mapping from image filenames to metadata indices."""
        self.image_to_metadata_index = dict()

        for idx, image_file in enumerate(self.image_files):
            image_filename = Path(image_file).name
            image_id = image_filename.replace("_merged.tif", "").replace(".tif", "")
            metadata_indices = self.metadata.index[self.metadata[self.col_name] == image_id].tolist()
            self.image_to_metadata_index[idx] = metadata_indices[0]

    def __len__(self) -> int:
        return len(self.image_files)

    def _get_date(self, row: pd.Series) -> torch.Tensor:
        """Extract and format temporal coordinates (T, date) from metadata."""
        temporal_coords = []
        for col in self.date_columns:
            date_str = row[col]
            date = pd.to_datetime(date_str, format="%Y-%m-%d")
            temporal_coords.append([date.year, date.dayofyear - 1])

        return torch.tensor(temporal_coords, dtype=torch.float32)

    def _get_coords(self, image: DataArray) -> torch.Tensor:
        px = image.x.shape[0] // 2
        py = image.y.shape[0] // 2

        # get center point to reproject to lat/lon
        point = image.isel(band=0, x=slice(px, px + 1), y=slice(py, py + 1))
        point = point.rio.reproject("epsg:4326")

        lat_lon = np.asarray([point.y[0], point.x[0]])

        return torch.tensor(lat_lon, dtype=torch.float32)

    def __getitem__(self, index: int) -> dict[str, Any]:
        image = self._load_file(self.image_files[index], nan_replace=self.no_data_replace)

        location_coords, temporal_coords = None, None
        if self.use_metadata:
            location_coords = self._get_coords(image)
            metadata_idx = self.image_to_metadata_index.get(index, None)
            if metadata_idx is not None:
                row = self.metadata.iloc[metadata_idx]
                temporal_coords = self._get_date(row)

        # to channels last
        image = image.to_numpy()
        if self.expand_temporal_dimension:
            image = rearrange(image, "(channels time) h w -> channels time h w", channels=len(self.bands))
        image = np.moveaxis(image, 0, -1)

        # filter bands
        image = image[..., self.band_indices]

        output = {
            "image": image.astype(np.float32),
            "mask": self._load_file(
                self.segmentation_mask_files[index], nan_replace=self.no_label_replace).to_numpy()[0],
        }

        if self.reduce_zero_label:
            output["mask"] -= 1
        if self.transform:
            output = self.transform(**output)
        output["mask"] = output["mask"].long()

        if self.use_metadata:
            output["location_coords"] = location_coords
            output["temporal_coords"] = temporal_coords

        return output

    def _load_file(self, path: Path, nan_replace: int | float | None = None) -> DataArray:
        data = rioxarray.open_rasterio(path, masked=True)
        if nan_replace is not None:
            data = data.fillna(nan_replace)
        return data

    def plot(self, sample: dict[str, Tensor], suptitle: str | None = None) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            suptitle: optional string to use as a suptitle

        Returns:
            a matplotlib Figure with the rendered sample
        """
        num_images = self.time_steps + 2

        rgb_indices = [self.bands.index(band) for band in self.rgb_bands]
        if len(rgb_indices) != 3:
            msg = "Dataset doesn't contain some of the RGB bands"
            raise ValueError(msg)

        images = sample["image"]
        images = images[rgb_indices, ...]  # Shape: (T, 3, H, W)

        processed_images = []
        for t in range(self.time_steps):
            img = images[t]
            img = img.permute(1, 2, 0)
            img = img.numpy()
            img = clip_image(img)
            processed_images.append(img)

        mask = sample["mask"].numpy()
        if "prediction" in sample:
            num_images += 1
        fig, ax = plt.subplots(1, num_images, figsize=(12, 5), layout="compressed")
        ax[0].axis("off")

        norm = mpl.colors.Normalize(vmin=0, vmax=self.num_classes - 1)
        for i, img in enumerate(processed_images):
            ax[i + 1].axis("off")
            ax[i + 1].title.set_text(f"T{i}")
            ax[i + 1].imshow(img)

        ax[self.time_steps + 1].axis("off")
        ax[self.time_steps + 1].title.set_text("Ground Truth Mask")
        ax[self.time_steps + 1].imshow(mask, cmap="jet", norm=norm)

        if "prediction" in sample:
            prediction = sample["prediction"]
            ax[self.time_steps + 1].axis("off")
            ax[self.time_steps+2].title.set_text("Predicted Mask")
            ax[self.time_steps+2].imshow(prediction, cmap="jet", norm=norm)

        cmap = plt.get_cmap("jet")
        legend_data = [[i, cmap(norm(i)), self.class_names[i]] for i in range(self.num_classes)]
        handles = [Rectangle((0, 0), 1, 1, color=tuple(v for v in c)) for k, c, n in legend_data]
        labels = [n for k, c, n in legend_data]
        ax[0].legend(handles, labels, loc="center")

        if suptitle is not None:
            plt.suptitle(suptitle)

        return fig



from collections.abc import Sequence
from typing import Any

import albumentations as A

from terratorch.datamodules.generic_multimodal_data_module import MultimodalNormalize
from terratorch.datamodules.generic_multimodal_data_module import wrap_in_compose_is_list
from terratorch.datasets import CarbonFluxNonGeo
from torchgeo.datamodules import NonGeoDataModule
from torchgeo.transforms import AugmentationSequential

MEANS = {
    "image": {
        "BLUE": 0.07372144372093026,
        "GREEN": 0.10117611215116282,
        "RED": 0.11269885680232558,
        "NIR": 0.2775572554069766,
        "SWIR_1": 0.21387001372093037,
        "SWIR_2": 0.14144541145348838
    },
    "merra_vars": [282.373169, 296.706468, 288.852922, 278.612209, 0.540145,
                   53.830276, 53.827718, 206.817980, 23.077581, 0.000003],
    "mask": 3.668982
}

STDS = {
    "image": {
        "BLUE": 0.13324302628303733,
        "GREEN": 0.13308921403475235,
        "RED": 0.13829909331863693,
        "NIR": 0.12039809083338567,
        "SWIR_1": 0.1088096350639653,
        "SWIR_2": 0.09366368859284444
    },
    "merra_vars": [9.296960, 11.402008, 10.311107, 8.064209, 0.171909,
                   49.945953, 48.907351, 74.591578, 8.746668, 0.000014],
    "mask": 3.804261
}


class CarbonFluxNonGeoDataModule(NonGeoDataModule):
    """NonGeo datamodule implementation for Landslide4Sense."""

    def __init__(
        self,
        data_root: str,
        batch_size: int = 4,
        num_workers: int = 0,
        bands: Sequence[str] = CarbonFluxNonGeo.all_band_names,
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        predict_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        no_data_replace: float | None = 0.0001,
        use_metadata: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init__(CarbonFluxNonGeo, batch_size, num_workers, **kwargs)
        self.data_root = data_root

        means = {
            m: ([MEANS[m][band] for band in bands] if m == "image" else MEANS[m])
            for m in MEANS.keys()
        }
        stds = {
            m: ([STDS[m][band] for band in bands] if m == "image" else STDS[m])
            for m in STDS.keys()
        }
        self.mask_means = MEANS["mask"]
        self.mask_std = STDS["mask"]
        self.bands = bands
        self.train_transform = wrap_in_compose_is_list(train_transform)
        self.val_transform = wrap_in_compose_is_list(val_transform)
        self.test_transform = wrap_in_compose_is_list(test_transform)
        self.predict_transform = wrap_in_compose_is_list(predict_transform)
        self.aug = MultimodalNormalize(means, stds) if aug is None else aug
        self.no_data_replace = no_data_replace
        self.use_metadata = use_metadata

    def setup(self, stage: str) -> None:
        if stage in ["fit"]:
            self.train_dataset = self.dataset_class(
                split="train",
                data_root=self.data_root,
                transform=self.train_transform,
                bands=self.bands,
                gpp_mean=self.mask_means,
                gpp_std=self.mask_std,
                no_data_replace=self.no_data_replace,
                use_metadata=self.use_metadata,
            )
        if stage in ["fit", "validate"]:
            self.val_dataset = self.dataset_class(
                split="test",
                data_root=self.data_root,
                transform=self.val_transform,
                bands=self.bands,
                gpp_mean=self.mask_means,
                gpp_std=self.mask_std,
                no_data_replace=self.no_data_replace,
                use_metadata=self.use_metadata,
            )
        if stage in ["test"]:
            self.test_dataset = self.dataset_class(
                split="test",
                data_root=self.data_root,
                transform=self.test_transform,
                bands=self.bands,
                gpp_mean=self.mask_means,
                gpp_std=self.mask_std,
                no_data_replace=self.no_data_replace,
                use_metadata=self.use_metadata,
            )
        if stage in ["predict"]:
            self.predict_dataset = self.dataset_class(
                split="test",
                data_root=self.data_root,
                transform=self.predict_transform,
                bands=self.bands,
                gpp_mean=self.mask_means,
                gpp_std=self.mask_std,
                no_data_replace=self.no_data_replace,
                use_metadata=self.use_metadata,
            )



from collections.abc import Sequence
from typing import Any

import albumentations as A
from torchgeo.transforms import AugmentationSequential

from terratorch.datamodules.geobench_data_module import GeobenchDataModule
from terratorch.datasets import MChesapeakeLandcoverNonGeo

MEANS = {"BLUE": 0.4807923436164856, "GREEN": 0.5200885534286499, "NIR": 0.569856584072113, "RED": 0.4570387601852417}

STDS = {"BLUE": 0.17441707849502563, "GREEN": 0.1976749747991562, "NIR": 0.2831788957118988, "RED": 0.21191735565662384}


class MChesapeakeLandcoverNonGeoDataModule(GeobenchDataModule):
    def __init__(
        self,
        batch_size: int = 8,
        num_workers: int = 0,
        data_root: str = "./",
        bands: Sequence[str] | None = None,
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        partition: str = "default",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            MChesapeakeLandcoverNonGeo,
            MEANS,
            STDS,
            batch_size=batch_size,
            num_workers=num_workers,
            data_root=data_root,
            bands=bands,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            aug=aug,
            partition=partition,
            **kwargs,
        )
